# -*- coding: utf-8 -*-
"""

Run script to analyse ESM residual data together with a storage


DFG SPP 2403

Created on Sun Jan 26 10:16:56 2025

@author: atakan
"""



import pandas as pd

import rws_py

from rws_py import RESULTS_DIR, app_suf, esm_residual_to_df

if __name__ == "__main__":
    COUNTRIES = ["DE0 Tot", "IT0 0", "GB5 0", "ES0 0", "IE4 0"]
    DIR_NAME = RESULTS_DIR
    input_dict = {"storage-hours-max" : 12,
                  }
    
    for act_country in COUNTRIES:
        f_name = f"{act_country}, {input_dict['storage-hours-max']} h"
    
    
    
        fil_name = app_suf(DIR_NAME,  f_name.replace(" ", "-").replace(",", ""))
        data_act = esm_residual_to_df(country=act_country)
        data_act.index = pd.to_datetime(data_act.index)
    
        # for storage sizing
        negative_values = data_act[data_act["residual-no-storage"] < 0]
        mean_negative_values = negative_values["residual-no-storage"].mean()
        st_size = input_dict["storage-hours-max"] * abs(mean_negative_values)
    
        # generate data with storage included
        data_act = rws_py.weather_load_storage(data_act,
                                        None,
                                        storage_size=st_size,
                                        residual=True)
        # get period lengths, mean values, following period with pos residuals
        eval_results = rws_py.residual_analyse(
            data_act, "residual-no-storage", thresholds={'direct': 0.})
    
        rws_py.plot_analysis_b(eval_results, file=fil_name, label=f_name)
        rws_py.plot_res(data_act, file=fil_name, label=f_name)
    
        storage_states, level_of_autonomy, charging = \
            rws_py.posterior_analysis_residual(data_act, storage_capacity=st_size)
    
        print("\n", f_name)
        rws_py.print_dict([storage_states, level_of_autonomy, charging],
                       app_suf(fil_name,"-stor-eval"))
    
        data_act.to_csv(app_suf(fil_name,"-time-series.csv"), sep=";")
        rws_py.plot_histo_analysis(eval_results, en_t =f_name, name_file=fil_name)
        rws_py.plot_histo_int_analysis(eval_results, en_t =f_name, name_file=fil_name,
                                    integral =True)







# -*- coding: utf-8 -*-
"""
New residual analysis unifying for weather data and ESM residuals.

Created on Sat Jan 25 11:21:18 2025

@author: atakan
"""

import json
import numpy as np
import pandas as pd
import seaborn as sbn
import matplotlib.pyplot as plt
from rws_py import app_suf


pd.set_option('display.max_columns', None)



def rename_count_columns(df):
    """
    Replace Column names in pandas dataFrame

    from count to period and from sum to integral, should help understanding/
    readability.

    Parameters
    ----------
    df : pandas DataFrame
        DtaFrame with the columns to replace.

    Returns
    -------
    pandas DataFrame
        column names changed.

    """
    return df.rename(columns={
        col: col.replace('count', 'period').replace('sum', 'integral')
        for col in df.columns
        if 'count' in col or 'sum' in col
    })




def residual_analyse(data, column, thresholds={'direct': 0.},
                     verbose=False, time_base_e=1.0):
    """
    Comprehensive analysis of time series data with threshold-based period analysis

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing the data to be analyzed
    column : str
        Name of the column to analyze
    thresholds : dict or None, optional
        Dictionary of thresholds for different analyses (example {"name":0.0}).
        Default is {'direct': 0.}.
    verbose : bool, optional
        If True, print additional information (default is False)
    time_base_e : float
        units in hours for determination of periods etc. default is 1.0

    Returns
    -------
    dict of analysis results
    """
    # Validate datetime index
    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("DataFrame must have a datetime index")

    # Calculate time resolution in hours
    time_deltas = data.index.to_series().diff()
    time_per_point = time_deltas.median().total_seconds() / 3600
    time_conversion = time_per_point / time_base_e

    # Create a copy of the data to avoid modifying the original
    # data = data.copy()

    # Initialize results dictionary
    results = {'time_resolution_hours': time_per_point}

    # Default analysis with zero threshold
    data['above_threshold'] = data[column] > 0
    data['block'] = (data['above_threshold'].diff() != 0).cumsum()

    # Period statistics
    df_period = data.groupby('block').agg(
        # Verwende die variable column
        count=(column, lambda x: len(x) * time_conversion),
        mean=(column, 'mean'),
        std=(column, 'std'),
        sum=(column, lambda x: sum(x) * time_conversion),
        min=(column, 'min'),
        max=(column, 'max'),
        time=(column, lambda x: x.index[0])  # Startzeitpunkt der Periode
    ).reset_index()

    # Add quantiles
    quantile_cols = data.groupby('block')[column].quantile([
        0.25, 0.5, 0.75]).unstack()
    df_period['25%'] = quantile_cols[0.25].values
    df_period['50%'] = quantile_cols[0.5].values
    df_period['75%'] = quantile_cols[0.75].values

    df_period['positive'] = df_period['sum'] > 0

    # Combinations of periods
    pos_neg_list, neg_pos_list = [], []
    for i in range(len(df_period)):
        current_row = df_period.iloc[i]
        if current_row['positive']:
            # Positive-Negative combination
            if i + 1 < len(df_period) and not df_period.iloc[i + 1]['positive']:
                next_row = df_period.iloc[i + 1]
                combined = pd.concat([current_row.add_suffix(
                    '_pos'), next_row.add_suffix('_neg')])
            else:
                zero_row = pd.Series(
                    0, index=current_row.index).add_suffix('_neg')
                combined = pd.concat(
                    [current_row.add_suffix('_pos'), zero_row])
            pos_neg_list.append(combined)
        else:
            # Negative-Positive combination
            if i + 1 < len(df_period) and df_period.iloc[i + 1]['positive']:
                next_row = df_period.iloc[i + 1]
                combined = pd.concat([current_row.add_suffix(
                    '_neg'), next_row.add_suffix('_pos')])
            else:
                zero_row = pd.Series(
                    0, index=current_row.index).add_suffix('_pos')
                combined = pd.concat(
                    [current_row.add_suffix('_neg'), zero_row])
            neg_pos_list.append(combined)

    df_pos_neg = pd.DataFrame(pos_neg_list).reset_index(drop=True)
    df_neg_pos = pd.DataFrame(neg_pos_list).reset_index(
        drop=True)  # This is the important one!

    # Verbose output
    if verbose:
        print("Positive-Negative Periods:")
        print(df_pos_neg)
        print("\nNegative-Positive Periods:")
        print(df_neg_pos)

    # Default analysis results and renaming
    results['default_analysis'] = {
        'period_stats': rename_count_columns(df_period),
        'pos_neg_periods': rename_count_columns(df_pos_neg),
        'neg-pos-res': rename_count_columns(df_neg_pos)

    }

    # Additional threshold analyses
    if thresholds:
        for threshold_name, threshold_value in thresholds.items():
            # Analyze periods above specific thresholds
            data['neg-residual'] = data[column] < threshold_value
            data['block_custom'] = (data['neg-residual'].diff() != 0).cumsum()

            # Threshold-specific analysis
            df_threshold = data.groupby('block_custom').agg(
                count=(column, lambda x: len(x) * time_conversion),
                mean=(column, 'mean'),
                std=(column, 'std'),
                sum=(column, lambda x: sum(x) * time_conversion),
                min=(column, 'min'),
                max=(column, 'max'),
                # Startzeitpunkt der Periode
                time=(column, lambda x: x.index[0])
            ).reset_index()

            # Storage period analysis
            neg_pos_residual_list = []
            for i in range(len(df_threshold)):
                if i + 1 < len(df_threshold):
                    current_row = df_threshold.iloc[i].copy()
                    next_row = df_threshold.iloc[i + 1]

                    if current_row['mean'] < threshold_value and \
                            next_row['mean'] >= threshold_value:
                        current_row['period-pos'] = next_row['count'].copy()
                        neg_pos_residual_list.append(current_row)

            df_neg_pos_residual = pd.DataFrame(neg_pos_residual_list)

            # Storage periods statistics
            storage_periods = {
                'count': len(df_neg_pos_residual),
                'mean_duration': df_neg_pos_residual['count'].mean() if not
                df_neg_pos_residual.empty else 0,
                'total_duration': df_neg_pos_residual['count'].sum() if not
                df_neg_pos_residual.empty else 0,
                'mean_intensity': df_neg_pos_residual['mean'].mean() if not
                df_neg_pos_residual.empty else 0,
                'mean_pos_duration': df_neg_pos_residual['period-pos'].mean()
                if not df_neg_pos_residual.empty else 0,
            }

            # Store results for this threshold
            results[f'{threshold_name}_threshold_analysis'] = {
                'period_data': rename_count_columns(df_threshold),
                'storage_periods': storage_periods,
                'neg-pos-res': rename_count_columns(df_neg_pos_residual)
            }

    return results


def posterior_analysis_residual(actual_df, storage_capacity=1., verbose=True):
    """
    Evaluate the residual load with energy storage DataFrame after its calculation

    The number of full charging/discharging cycles ("charging"), the level of
    autonomy, and the times the storag changed its state either from empty or
    from full("storage_states are evaluated.

    Parameters
    ----------
    actual_df : panda DataFrame
        the data timeseries with the residuals, storage states.
    storage_capacity : Float, optional
        max. storage capacity, needed to calculate the equivalent full charging
        /discharging cycles. The default is 1..
    verbose : Boolean, optional
        for getting some printings. The default is True.

    Returns
    -------
    storage_states : Dictionary
        DESCRIPTION.
    level_of_autonomy : Dictionary
        DESCRIPTION.
    charging : Dictionary
        DESCRIPTION.

    """

    number_all = actual_df['SOC'].count()
    soc_counts = actual_df['SOC'].value_counts()
    zero_count = soc_counts.get(0, 0)  # Anzahl der 0en
    one_count = soc_counts.get(1, 0)    # Anzahl der 1en
    if "time_diff_seconds" in actual_df.columns:
        power_energy = actual_df["time_diff_seconds"].mean() / 3600
    else:
        print("WARNING: Time difference not n dataframe, assumed to be 1 h!")
        power_energy = 1.0

    # Berechnung der Differenzen zwischen aufeinanderfolgenden Werten
    actual_df['previous_SOC'] = actual_df['SOC'].shift(
        1)  # Vorheriger Wert

    # Bedingung für Wechsel von 0 auf einen höheren Wert
    from_0_to_higher = ((actual_df['previous_SOC'] == 0) & (
        actual_df['SOC'] > 0)).sum()

    # Bedingung für Wechsel von 1 auf einen niedrigeren Wert
    from_1_to_lower = ((actual_df['previous_SOC'] == 1) & (
        actual_df['SOC'] < 1)).sum()

    # Ausgabe der Ergebnisse
    if verbose:

        print(f"""Anzahl der Werte 0 in 'SOC': {
              zero_count} und 1 in 'SOC': {one_count}""")
        print(f"""Anzahl der Wechsel von 0 auf einen höheren Wert: {
              from_0_to_higher}; von 1 auf einen niedrigeren Wert: {
            from_1_to_lower}""")

    storage_states = {
        "empty": zero_count / number_all,
        "full": one_count / number_all,
        "charging": from_0_to_higher,
        "discharging":  from_1_to_lower
    }
    c_name = "residual-no-storage"
    total_pos_residual = (actual_df.loc[actual_df[c_name] > 0, c_name] *
                          power_energy).sum()
    total_neg_residual = (actual_df.loc[actual_df[c_name] < 0, c_name] *
                          power_energy).sum()
    residual_ratio_n_p = abs(total_neg_residual / total_pos_residual)

    c_name = "residual-w-storage"
    total_pos_residual_w = (actual_df.loc[actual_df[c_name] > 0, c_name] *
                          power_energy).sum()
    total_neg_residual_w = (actual_df.loc[actual_df[c_name] < 0, c_name] *
                          power_energy).sum()
    residual_ratio_n_p_w = abs(total_neg_residual_w / total_pos_residual_w)
    # Zugrunde liegende Werte zur LA Auswertung
    time_pos_res_remain = actual_df[actual_df[c_name] > 0].count()[c_name]
    time_neg_res_remain = actual_df[actual_df[c_name] < 0].count()[c_name]
    time_no_res_remain = actual_df[actual_df[c_name] == 0].count()[c_name]

    la_pos = 1 - (time_pos_res_remain / actual_df[c_name].count())
    la_neg = 1 - (time_neg_res_remain / actual_df[c_name].count())
    la_eq = time_no_res_remain / actual_df[c_name].count()
    level_of_autonomy = {
        "LA-pos": la_pos,
        "LA-neg": la_neg,
        "Zero-Residual-Time": la_eq,
        "Pos. Residuals (no-storage)" : total_pos_residual,
        "Neg. Residuals (no-storage)" : total_neg_residual,
        "Neg./Pos. residual ratio (no-storage)" : residual_ratio_n_p,
        "Pos. Residuals (w-storage)" : total_pos_residual_w,
        "Neg. Residuals (w-storage)" : total_neg_residual_w,
        "Neg./Pos. residual ratio (w-storage)" : residual_ratio_n_p_w,
        
        
    }

    charging = None
    c_name = "power_st_in"
   
    total_energy_to_storage = (
        actual_df.loc[actual_df[c_name] > 0, c_name] *
        power_energy).sum()

    # Skalieren der negativen Werte (Energie aus dem Speicher entnommen)
    total_energy_from_storage = (
        actual_df.loc[actual_df[c_name] < 0, c_name] *
        power_energy).sum()
   
    number_charging = total_energy_to_storage / storage_capacity
    number_discharging = total_energy_from_storage / storage_capacity

    charging = {"total energy stored": total_energy_to_storage,
                "total energy discharged": total_energy_from_storage,
                "Number of full charging": number_charging,
                "Number of full discharging": number_discharging,

                }

    return storage_states, level_of_autonomy, charging


def _print_dict(my_dict):
    for k, v in my_dict.items():
        if isinstance(v, np.float64):
            print(f"{k}: {float(v):.4f}")
        elif isinstance(v, float):
            print(f"{k}: {v:.4f}")
        else:
            print(f"{k}: {v}")


def print_dict(my_dict, filename=None):
    """
    print the dictionary and optional: store it to the json file

    Parameters
    ----------
    my_dict : dictionary
        DESCRIPTION.
    filename : string, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    output : TYPE
        DESCRIPTION.

    """

    # Funktion zum Verarbeiten eines einzelnen Dictionaries
    def process_single_dict(d):
        output = {}
        for k, v in d.items():
            if isinstance(v, (np.float64, float)):
                output[k] = f"{float(v):.4f}"
            elif isinstance(v, (np.int64, float)):
                output[k] = f"{int(v)}"
            else:
                output[k] = v
            print(f"{k}: {output[k]}")
        return output

    # Wenn my_dict eine Liste von Dictionaries ist
    if isinstance(my_dict, list):
        processed_list = [process_single_dict(d) for d in my_dict]
        output_to_save = processed_list
    else:
        # Ansonsten wird das einzelne Dictionary verarbeitet
        output_to_save = process_single_dict(my_dict)

    # Speichern in einer JSON-Datei, wenn ein Dateiname angegeben ist
    if filename is not None:
        try:
            with open(app_suf(filename, ".json"), 'w') as f:
                json.dump(output_to_save, f, indent=4)
            print(
                f"Ergebnisse wurden in {app_suf(filename,'.json')} gespeichert.")
        except Exception as e:
            print(f"Fehler beim Schreiben in die Datei '{filename}': {e}")


def plot_analysis_b(evaluation_df, thresh_n="default", file=None, label=None,
                    y_max=None):
    """
    Scatterplot of periods with neg. residuals vs. the following pos. residual

    Parameters
    ----------
    evaluation_df : dictionary withpandas dataFrame
        period lengths and integgrals for negative residuals and th directly
        following positive residuals.
    thresh_n : float, optional
        threshold-dictionary name (part, typically "basic", "direct" or
                                   "default". The default is "default".
    file : string, optional
        filename without ending for storing the plot. The default is None.
    label : string, optional
        title of the plot. The default is None.

    Returns
    -------
    None.

    """

    plt.figure()

    # neg_pos[['integral_pos', 'integral_neg']] = neg_pos[[
    #     'integral_pos', 'integral_neg']] * SUM_UNITS
    if thresh_n == "default":
        direct = evaluation_df[thresh_n + '_analysis']
        df_act = direct["neg-pos-res"]
        sbn.scatterplot(data=df_act,
                        x='period_neg', y='period_pos',
                        size='integral_neg', hue='integral_pos')
    else:
        direct = evaluation_df[thresh_n + '_threshold_analysis']
        df_act = direct["neg-pos-res"]
        sbn.scatterplot(data=df_act,
                        x='period', y='period-pos',
                        size='integral', hue='mean')  # , style='mean_pos')
    if y_max is not None:
        plt.ylim(0, y_max)
    plt.legend(loc='upper left', bbox_to_anchor=(1, 1))

    # Optional: Ändere die Größe des Plots, um Platz für die Legende zu schaffen
    plt.tight_layout()
    plt.title(label)
    plt.savefig(app_suf(file, "_scat_neg_pos", "jpg"))

    plt.show()
    if file is not None:
        df_act.to_csv(app_suf(file, "-eval-neg-pos-res", "csv"))
        print_dict(evaluation_df['direct_threshold_analysis']
                   ["storage_periods"], app_suf(file, "-eval-st-periods"))


def plot_res(dat_in,  file=None, label=None):
    """
    Plot time dependent residuals with and w/o storage, loads and power

    in case of ESM residual analysis, without öoad and power

    Parameters
    ----------
    dat_in : DataFrame
        all data, index is datetime.
    file : string, optional
        filename without ending for storing the plot. The default is None.
    label : string, optional
        plot title. The default is None.

    Returns
    -------
    None.

    """
    n_s = 2
    with_power = "power" in dat_in.columns
    if with_power:
        n_s += 1
    figx, ax = plt.subplots(n_s, sharex=True, figsize=(
        10, 8), constrained_layout=True)  # Größere Abbildung für Platz
    if with_power:
        dat_in.plot(y="power", ax=ax[n_s-1], legend=True)
        dat_in.plot(y="load", ax=ax[n_s-1], legend=True)
    dat_in.plot(y="residual-w-storage", ax=ax[0], legend=True)
    dat_in.plot(y="residual-no-storage", ax=ax[1], legend=True)
    dat_in.plot(y="act_st_cap", ax=ax[1], legend=True)

    # Legenden außerhalb platzieren
    for axis in ax:
        axis.legend(loc="upper left", bbox_to_anchor=(
            1.02, 1))  # Position außerhalb rechts

    ax[0].set_title(label)
    # Zusätzlicher Platz bei Speicherung
    figx.savefig(app_suf(file, "_res_storage_time", "jpg"),
                 bbox_inches="tight")
    plt.show()


def generate_test_data(duration_hours=48, resolution_minutes=60):
    """Generate test data with datetime index"""
    timestamps = pd.date_range(
        start='2024-01-01',
        periods=int(duration_hours * 60 / resolution_minutes),
        freq=f'{resolution_minutes}min'
    )

    # Generate sine wave
    time_array = np.linspace(0, duration_hours, len(timestamps))
    sine_values = np.sin(2 * np.pi * (1/12) * time_array)

    return pd.DataFrame({'value': sine_values}, index=timestamps)


def plot_histo_analysis(results, select=None,  en_t="title", name_file="histo",
                        bins=[0, 1, 2, 4, 8, 16, 32, 64, 128, 200],
                        integral=False,
                        n_bins_middle=10):
    """
    Create  visualizations for the analysis results with improved period distribution plots
    """
    pl_name = f"{en_t}"
    # Visualize results
    if select is None:
        select = {'lev-1': "default_analysis",
                  'lev-2': "neg-pos-res",
                  'col-1': 'period_neg',
                  'col-2': 'period_pos',
                  'col-1a': 'integral_neg',
                  'col-2a': 'integral_pos',

                  }

    act_df = results[select['lev-1']][select['lev-2']]
    if integral:
        bins_all = []
        for dat_bin in [act_df[select['col-1a']], act_df[select['col-2a']]]:
            p10 = np.percentile(dat_bin, 10)
            p90 = np.percentile(dat_bin, 90)
            print(p90, p10)
            # Create bin edges
            middle_bins = np.linspace(p10, p90, n_bins_middle + 1)
            bins_all.append(np.concatenate(
                [[dat_bin.min()], middle_bins, [dat_bin.max()]]))
        histo_2, xedges, yedges = np.histogram2d(
            act_df[select['col-1a']], act_df[select['col-2a']], bins=bins_all)
    else:

        histo_2, xedges, yedges = np.histogram2d(
            act_df[select['col-1']], act_df[select['col-2']], bins=bins)

    fig0, ax0 = plt.subplots(1)
    histo_2=histo_2.T

    ax0.imshow(histo_2)
    xranges = [
        f"neg {xedges[i]:.1f} - {xedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    yranges = [
        f"pos {yedges[i]:.1f} - {yedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    ax0.set_xticks(np.arange(len(xranges)), labels=xranges,
                   rotation=45, ha="right", rotation_mode="anchor")
    ax0.set_yticks(np.arange(len(yranges)), labels=yranges)
    for i in range(len(yranges)):
        for j in range(len(xranges)):
            if histo_2[i, j] > 0:
                ax0.text(j, i, f"{histo_2[i, j]:.0f}",
                         ha="center", va="center", color="w", fontsize=10)

    ax0.set_title(pl_name+", 2D Histogram")
    fig0.tight_layout()
    fig0.savefig(app_suf(name_file, "-2D-hist.png"))
    plt.show()
    df = pd.DataFrame(histo_2.T, index=yranges, columns=xranges)

    # xedges als eigene Spalte hinzufügen
    df.insert(0, "y_edges_start", yedges[:-1])

    # CSV speichern
    df.to_csv(app_suf(name_file, "histogram_data.csv"))

    # fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(ncols=2, nrows=2)

    # pcm = ax1.pcolormesh(xedges, yedges, histo_2, cmap='rainbow')
    # fig.colorbar(pcm, ax=ax1)
    # ax1.set_xlabel("Period length / h")
    # ax1.set_ylabel("pause length / h")
    # sum_count_fit = Polynomial.fit(
    #     results['direct_inactive']['count'], results['direct_inactive']['sum'], 2)
    # ax2.plot(results['direct_inactive']['count'],
    #          results['direct_inactive']['sum'], "+k")
    # ax2.plot(results['direct_inactive']['count'], sum_count_fit(
    #     results['direct_inactive']['count']), '.')
    # ax2.set_xlabel("Period length / h")
    # ax2.set_ylabel("total intensity")
    # ax3.plot(period_pause['count'], period_pause['mean'], "+k")
    # ax3.set_xlabel("Period length / h")
    # ax3.set_ylabel("mean intensity")


def plot_histo_int_analysis(results, select=None,  en_t="title", name_file="histo",
                            bins=[0, 1, 2, 4, 8, 16, 32, 64, 128, 200],
                            integral=False,
                            n_bins_middle=5,
                            dict_dict=True):
    """
    Create  visualizations for the analysis results with improved period distribution plots
    """
    pl_name = f"{en_t}"
    # Visualize results
    if select is None:
        select = {'lev-1': "default_analysis",
                  'lev-2': "neg-pos-res",
                  'col-1': 'period_neg',
                  'col-2': 'period_pos',
                  'col-1a': 'integral_neg',
                  'col-2a': 'integral_pos',

                  }
    if dict_dict:
        act_df = results[select['lev-1']][select['lev-2']]
    else:
        act_df = results
    if integral:
        bins_all = []
        for i_d, dat_bin in enumerate([act_df[select['col-1a']], act_df[select['col-2a']]]):
            p10 = np.percentile(dat_bin, 10)
            p90 = np.percentile(dat_bin, 90)
            if i_d==0:
                p90=-1
            elif i_d==1:
                p10 =1
            print(p90, p10)
            # Create bin edges
            middle_bins = np.linspace(p10, p90, n_bins_middle + 1)
            bins_all.append(np.concatenate(
                [[dat_bin.min()], middle_bins, [dat_bin.max()]]))
        histo_2, xedges, yedges = np.histogram2d(
            act_df[select['col-1a']], act_df[select['col-2a']], bins=bins_all)
    else:

        histo_2, xedges, yedges = np.histogram2d(
            act_df[select['col-1']], act_df[select['col-2']], bins=bins)

    histo_2=histo_2.T
    x_hist = np.histogram(act_df[select['col-1a']], bins=xedges)
    y_hist = np.histogram(act_df[select['col-2a']], bins=yedges)
    hist_2n = np.zeros((n_bins_middle+3, n_bins_middle+3))
    hist_2n[:-1, :-1]=histo_2
    hist_2n[:-1, -1]=y_hist[0]
    hist_2n[-1, :-1]=x_hist[0]
    histo_2 = hist_2n
    fig0, ax0 = plt.subplots(1, figsize=(
        10, 8), constrained_layout=True)
    

    ax0.imshow(histo_2)
    xranges = [
        f"{xedges[i]:.1f} - {xedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    yranges = [
        f"{yedges[i]:.1f} - {yedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    xranges.append("sum")
    yranges.append("sum")
    ax0.set_xticks(np.arange(len(xranges)), labels=xranges,
                   rotation=45, ha="right", rotation_mode="anchor")
    ax0.set_yticks(np.arange(len(yranges)), labels=yranges)
    for i in range(len(yranges)):
        for j in range(len(xranges)):
            if histo_2[i, j] > 0:
                ax0.text(j, i, f"{histo_2[i, j]:.0f}",
                         ha="center", va="center", color="w", fontsize=10)

    ax0.set_title(pl_name+", 2D Int-Histogram")
    fig0.tight_layout()
    fig0.savefig(app_suf(name_file, "-2D-hist-int.png"))
    plt.show()
    df = pd.DataFrame(histo_2.T, index=yranges, columns=xranges)

    # xedges als eigene Spalte hinzufügen
    df.insert(0, "y_edges_start", yedges[:-1])

    # CSV speichern
    df.to_csv(app_suf(name_file, "integral-histogram_data.csv"))


if __name__ == "__main__":

    data_out = generate_test_data(duration_hours=48, resolution_minutes=60)

    # Analyze with different thresholds
    result = residual_analyse(
        data_out, 'value',
        thresholds={'basic': -0.1, 'direct': 0.}
    )
    print("\n", result["basic_threshold_analysis"]["period_data"])
    fig, (ax1, ax2) = plt.subplots(2)
    sbn.scatterplot(data=result["basic_threshold_analysis"]
                    ["neg-pos-res"], x="period", y="period-pos", ax=ax1)
    sbn.scatterplot(data=result["direct_threshold_analysis"]
                    ["neg-pos-res"], x="period", y="period-pos", ax=ax1)
    data_out.plot(y="value", ax=ax2)

    # Basic sign-based analysis
    result1 = residual_analyse(data_out, 'value')
    print("\n", result1["default_analysis"])
    # # With basic threshold
    # result2 = residual_analyse(data, 'value',
    #     thresholds={'basic': 10})

    # # With basic and direct thresholds
    # result3 = residual_analyse(data, 'value',
    #     thresholds={

    #         'direct': 0,

    #     })



# -*- coding: utf-8 -*-
"""
Created on Sat Feb  1 10:59:13 2025

@author: atakan
"""
import pytest
import pandas as pd
import numpy as np
from rws_py.load_curve_simple import generate_load_curve

def test_generate_load_curve():
    # Testdaten
    time_ind = pd.date_range(start='2021-01-01', end='2022-12-31', freq="h")
    load_params = {
        'min_weekday': 0.650,
        'max_weekday': 1.0,
        'min_weekend': 0.650,
        'max_weekend': 0.85,
        'intermediate_fraction': 0.80,
        'multiplier': 50.0,
        'offset': 0.0,
        'normalize': True,
        'amplitude_season': 0.05,  # seasonal
    }

    # Generiere die Lastkurven
    load_curve_step = generate_load_curve(time_ind, step=True, load_parameters=load_params)
    load_curve_sine = generate_load_curve(time_ind, step=False, load_parameters=load_params)

    # Überprüfe die Summe/Integral der Lastkurven
    int_step = load_curve_step["load"].sum()
    int_sine = load_curve_sine["load"].sum()

    assert np.isclose(int_step, 17497.0), f"Expected 17497.0, but got {int_step}"
    assert np.isclose(int_sine, 17497.0), f"Expected 17497.0, but got {int_sine}"

    # Überprüfe die statistischen Werte der Lastkurven
    step_describe = load_curve_step.describe()
    sine_describe = load_curve_sine.describe()

   
    expected_step_describe = pd.Series({
        'count': 17497.0,
        'mean': 1.000000,
        'std': 0.177557,
        'min': 0.742097,
        '25%': 0.810880,
        '50%': 0.999602,
        '75%': 1.166246,
        'max': 1.261863
    })

    expected_sine_describe = pd.Series({
        'count': 17497.000000,
        'mean': 1.000000,
        'std': 0.149037,
        'min': 0.768476,
        '25%': 0.866095,
        '50%': 0.984161,
        '75%': 1.124193,
        'max': 1.306717
    })


    pd.testing.assert_series_equal(step_describe['load'], expected_step_describe, check_names=False)
    pd.testing.assert_series_equal(sine_describe['load'], expected_sine_describe, check_names=False)

if __name__ == '__main__':
    pytest.main()





# -*- coding: utf-8 -*-
"""

tests for the weather data from dws and the analysis
-not yet useful !

Created on Mon Feb  3 12:52:34 2025

@author: atakan
"""
import os
import unittest
import pandas as pd
from pathlib import Path
from rws_py import load_data, analyze_weather, resample_to_hourly, plot_analysis, plot_enhanced_analysis

class TestRWS(unittest.TestCase):

    def test_load_data(self):
        base_url = "https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/solar/historical"
        filename = "10minutenwerte_SOLAR_00891_20200101_20231231_hist.zip"
        expected_columns = ["MESS_DATUM", "QN", "GS_10"]
        data = load_data(base_url, filename, expected_columns)
        self.assertIsInstance(data, pd.DataFrame)

    def test_analyze_weather(self):
        # Beispielhafte Testdaten
        data = pd.DataFrame({
            "MESS_DATUM": pd.date_range(start="2020-01-01", periods=4, freq='H'),
            "power": [10, 20, 30, 40]
        }).set_index('MESS_DATUM')
        
        column = "power"
        threshold = 15
        
        # Aufruf der Funktion
        results = analyze_weather(data, column, threshold)
        
        # Überprüfen des Rückgabeformats
        self.assertIsInstance(results, tuple)
        self.assertEqual(len(results), 3)
        
        df_period, df_active_inactive, df_inactive_active = results
        
        # Weitere Assertions, um die Inhalte der DataFrames zu prüfen
        self.assertIsInstance(df_period, pd.DataFrame)
        self.assertIsInstance(df_active_inactive, pd.DataFrame)
        self.assertIsInstance(df_inactive_active, pd.DataFrame)
        
        # Angepasste Überprüfung der Spalten, einschließlich 'block'
        self.assertListEqual(
            list(df_period.columns),
            ['block', 'count', 'mean', 'std', 'sum', 'min', 'max']
        )
        self.assertListEqual(
            list(df_active_inactive.columns),
            ['count', 'mean', 'std', 'sum', 'min', 'max']
        )
        self.assertListEqual(
            list(df_inactive_active.columns),
            ['block','count', 'mean', 'std', 'sum', 'min', 'max', 'count_inv']
        )

    

import pytest
from unittest import mock
from unittest.mock import patch, MagicMock
import pandas as pd
from pathlib import Path
import shutil
import tempfile
import rws_py._weather_analysis as weather_analysis # Stellen Sie sicher, dass das Verzeichnis als Paket erkannt wird

@pytest.fixture
def mock_load_data():
    # Mock-Daten für Solar und Wind
    solar_data = pd.DataFrame({
        "MESS_DATUM": pd.date_range(start="2020-01-01", periods=24, freq='H'),
        "QN": [18]*24,
        "GS_10": [i * 10 for i in range(24)],
        "power": [i * 16.66 for i in range(24)]
    }).set_index("MESS_DATUM")

    wind_data = pd.DataFrame({
        "MESS_DATUM": pd.date_range(start="2020-01-01", periods=24, freq='H'),
        "QN": [3]*24,
        "FF_10": [i * 1.5 for i in range(24)],
        "power": [(i * 1.5) ** 3 for i in range(24)]
    }).set_index("MESS_DATUM")

    def load_data_side_effect(base_url, filename, expected_columns):
        if "SOLAR" in filename:
            return solar_data
        elif "wind" in filename:
            return wind_data
        else:
            return pd.DataFrame()

    with patch('weather_analysis.load_data', side_effect=load_data_side_effect):
        yield

@pytest.fixture
def temporary_results_dir():
    # Erstellen eines temporären Verzeichnisses für die Ergebnisse
    temp_dir = tempfile.mkdtemp()
    with patch('weather_analysis.RESULTS_DIR', Path(temp_dir)):
        yield Path(temp_dir)
    # Bereinigen nach dem Test
    shutil.rmtree(temp_dir)

@pytest.fixture
def mock_plot_functions():
    # Mocken der Plot-Funktionen, um das Erstellen von Dateien zu verhindern
    with patch('weather_analysis.plot_analysis') as mock_plot_analysis, \
         patch('weather_analysis.plot_enhanced_analysis') as mock_plot_enhanced:
        yield mock_plot_analysis, mock_plot_enhanced

def test_main_flow(mock_load_data, temporary_results_dir, mock_plot_functions):
    """
    Integrationstest für den Hauptablauf der main()-Funktion.
    """
    mock_plot_analysis, mock_plot_enhanced = mock_plot_functions

    # Mocken Sie andere abhängige Funktionen, falls notwendig
    with patch('weather_analysis.generate_load_curve', return_value=pd.DataFrame()), \
         patch('weather_analysis.weather_load_storage', return_value=pd.DataFrame()), \
         patch('weather_analysis.analyze_residual', return_value={
             'active_inactive': pd.DataFrame({
                 'count': [1, 2, 3],
                 'count_inv': [4, 5, 6],
                 'mean': [10, 20, 30],
                 'sum': [100, 200, 300]
             }),
             'basic_stats': {
                 'mean_active_periods': 15
             },
             'direct_threshold': 12.5,
             'storage_periods': {
                 'count': 2,
                 'mean_duration': 10,
                 'mean_pause_duration': 5,
                 'std_pause_duration': 2,
                 'std_duration': 3,
                 'sum': 100
             }
         }):
        # Führen Sie die main()-Funktion aus
       weather_analysis.main()

    # Überprüfen Sie, ob die Evaluierungs-CSV erstellt wurde
    eval_csv = temporary_results_dir / "H-all-eval-enhanced.csv"
    assert eval_csv.exists(), "Evaluierungs-CSV wurde nicht erstellt."

    # Laden und prüfen Sie den Inhalt der Evaluierungs-CSV
    evaluation_df = pd.read_csv(eval_csv, sep=';')
    assert not evaluation_df.empty, "Evaluierungs-CSV ist leer."
    expected_columns = ["Station", "Year", "power", "Threshold", "sum", "mean",
                        "10 min periods", "periods>1h", "sum with energy",
                        "sum without energy", "ratio of period lengths",
                        "mean_active_periods", "direct_threshold",
                        "storage_periods_count", "period_mean_duration", "mean_pause_duration",
                        'std_pause_duration', 'std_duration']
    for col in expected_columns:
        assert col in evaluation_df.columns, f"Spalte {col} fehlt in der Evaluierungs-CSV."

    # Weitere Überprüfungen können hinzugefügt werden, z.B. die Anzahl der Zeilen
    assert len(evaluation_df) > 0, "Evaluierungs-CSV enthält keine Daten."

    # Überprüfen Sie, ob Plot-Funktionen aufgerufen wurden
    mock_plot_analysis.assert_called()
    mock_plot_enhanced.assert_called()

    # Optional: Überprüfen Sie spezifische Werte in der Evaluierungs-CSV
    first_row = evaluation_df.iloc[0]
    assert first_row['Station'] == 'Cuxhaven', "Station-Name stimmt nicht."
    assert first_row['Year'] == 2020, "Jahr stimmt nicht."
    assert first_row['power'] == 752.447269, "Power-Wert stimmt nicht."
    # Fügen Sie weitere Assertions hinzu, um andere Werte zu überprüfen


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Create a HDF5 File out of a folder with subfolders

Created on Mon Feb 10 21:45:58 2025

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""

import os
import h5py
import yaml
import numpy as np
from pathlib import Path

def load_yaml(file_path):
    """Lädt eine YAML-Datei und gibt sie als Dictionary zurück."""
    with open(file_path, 'r', encoding='utf-8') as f:
        return yaml.safe_load(f)

def add_metadata(hdf5_obj, metadata):
    """Fügt Metadaten als Attribute zu einem HDF5-Objekt hinzu."""
    for key, value in metadata.items():
        if isinstance(value, (int, float, str, bool)):
            hdf5_obj.attrs[key] = value
        else:
            hdf5_obj.attrs[key] = str(value)  # Komplexe Daten als String speichern

def store_data_in_hdf5(hdf5_file, base_path, current_path):
    """Rekursives Speichern von Dateien und Metadaten in HDF5."""
    rel_path = os.path.relpath(current_path, base_path)
    group = hdf5_file.create_group(rel_path) if rel_path != '.' else hdf5_file
    
    yaml_metadata = {}
    
    # Suche nach YAML-Dateien
    for file in Path(current_path).glob("*.yaml"):
        yaml_metadata[file.stem] = load_yaml(file)
    
    # Falls es eine allgemeine YAML-Datei gibt, die nicht mit einer Datei übereinstimmt
    general_metadata = yaml_metadata.pop(Path(current_path).name, None)
    if general_metadata:
        add_metadata(group, general_metadata)
    
    for file in Path(current_path).iterdir():
        if file.is_dir():
            store_data_in_hdf5(hdf5_file, base_path, file)
        elif file.suffix not in ['.yaml', '.hdf5']:
            dataset_name = os.path.join(rel_path, file.name)
            try:
                data = np.loadtxt(file, delimiter=',') if file.suffix in ['.csv', '.txt'] else open(file, 'rb').read()
                dset = hdf5_file.create_dataset(dataset_name, data=data)
                
                # Falls es eine passende YAML-Datei gibt, Metadaten speichern
                if file.stem in yaml_metadata:
                    add_metadata(dset, yaml_metadata[file.stem])
            except Exception as e:
                print(f"Fehler beim Speichern von {file}: {e}")

def convert_folder_to_hdf5(source_folder, output_hdf5):
    """Konvertiert einen gesamten Ordner rekursiv in eine HDF5-Datei."""
    with h5py.File(output_hdf5, 'w') as hdf5_file:
        store_data_in_hdf5(hdf5_file, source_folder, source_folder)
    print(f"HDF5-Datei gespeichert unter: {output_hdf5}")

# Beispielaufruf:
# convert_folder_to_hdf5("mein_ordner", "output.hdf5")



# -*- coding: utf-8 -*-
"""
Read weather data, estimated load, calculate residuals, analyze pause etc.

Created on Wed Jan 22 12:43:27 2025

@author: atakan
"""
import pandas as pd
import rws_py
import rws_py.dwd_analyze_store_second_threshold as awls
from rws_py import generate_load_curve
# import residual_analyse as resa
#import rws_py.analysis_universal as anu
from rws_py import RESULTS_DIR, app_suf


if __name__ == "__main__":
    SUM_UNITS = 1.0
    HOURLY = "h" # None  # conversion to houtly data , or stick to 10 min?
    N_BEG = 0
    N_END = -1  # analyze data until to this index (for fast computation)
    CALC_ST = True
    DIR_NAME = RESULTS_DIR
    act_stations ={'Cuxhaven': '00891',}
    
    input_dict = {"load-scaling": 1.070,  # scales the load relative to the meaan value of the energy
                  # hours to calculate the storage capacity, multiplied with the max. value of
                  "storage-hours-max": 10.,
                  }
    
    
    if __name__ == "__main__":
        VERBOSE = True
    
        all_dat = []
        eval_docu=[]
        ENERGY_COLUMN = "power"  # Name of the column with the energy proportional data
    
        # Prepare data sources
        for ACT_STATION in act_stations.keys(): #awls.stations.keys():
            SOLAR_ACT = '10minutenwerte_SOLAR_' + \
                awls.stations[ACT_STATION]+'_20200101_20231231_hist.zip'
            WIND_ACT = '10minutenwerte_wind_' + \
                awls.stations[ACT_STATION]+'_20200101_20231231_hist.zip'
            all_dat.append([awls.SOLAR_URL_BASE, SOLAR_ACT,
                           awls.solar_columns, awls.SOLAR_THRESHOLD, ACT_STATION])
            all_dat.append([awls.WIND_URL_BASE, WIND_ACT, awls.wind_columns,
                           awls.WIND_THRESHOLD, ACT_STATION])
    
        # Main analysis loop
        for dat_act in all_dat:
            base_url = dat_act[0]
            filename = dat_act[1]
            expected_columns = dat_act[2]
            threshold = dat_act[3]
            ACT_STATION = dat_act[4]
    
            # Load data
            data_all = awls.load_data(base_url, filename, expected_columns,
                                      resample = HOURLY)
    
            # Process each year
            for year in range(2020, 2021):
    
    
                en_t = ACT_STATION+", "
                if dat_act[0].count('wind'):
                    ACTUAL_ENERGY = "Wind"
                else:
                    ACTUAL_ENERGY = "Solar"
                en_t += ACTUAL_ENERGY
                TITLE_NAME = f"{en_t}, {year}"
                print(f"\nNOW: {TITLE_NAME}")
    
                FILE_NAME = app_suf(DIR_NAME, TITLE_NAME.replace(", ", "-"))
                data = data_all[data_all.index.year == year]
                data_act = data.iloc[N_BEG: N_END]
    
                
    
                load_par = {
                    'min_weekday': 0.650,
                    'max_weekday': 1.0,
                    'min_weekend': 0.650,
                    'max_weekend': 0.85,  # only for step-function
                    'intermediate_fraction': .75,
                    # scales everything
                    'multiplier': 1.0,
                    'offset': 0.0,
                    'normalize': True,
                    'amplitude_season':0.05, 
                }
                load_df = generate_load_curve(
                    data_act.index, load_parameters=load_par)
                data_act =rws_py.calc_residuals(data_act, 
                                                load_df,
                                                gamma=input_dict["load-scaling"])
                # residuals calculated and in data_act now
                
                # let's analyse them
                eval_results = rws_py.residual_analyse(
                    data_act, "residual-no-storage")
                rws_py.plot_analysis_b(eval_results, file=FILE_NAME, label=TITLE_NAME)
                
                # here is the storage part now
                
                
    
                if CALC_ST is True: # analyse residuals with storage
                    print("Berechnung von st_size:",
                          input_dict["storage-hours-max"] * data_act["power"].mean())
                    st_size = (input_dict["storage-hours-max"]
                               * data_act["power"].mean()).copy()
                    print("Neuer Wert von st_size:", st_size)
                with_load = rws_py.weather_load_storage(data_act,
                                                      load_df,
                                                      storage_size=st_size,
                                                      residual =True,
                                                      )
    
    
                # Perform period analysis (includes basic analysis)
    
                data_act = with_load.copy()
                
    
                
                rws_py.plot_res(data_act, file=FILE_NAME, label=TITLE_NAME)
    
                storage_states, level_of_autonomy, charging = rws_py.posterior_analysis_residual(
                    data_act, storage_capacity=st_size)
    
                print("\n", TITLE_NAME)
                rws_py.print_dict([storage_states, level_of_autonomy,
                               charging], filename=app_suf(FILE_NAME,"eval-storage"))
                rws_py.plot_histo_analysis(
                    eval_results, en_t=TITLE_NAME, name_file=FILE_NAME)
                rws_py.plot_histo_int_analysis(
                    eval_results, en_t=TITLE_NAME, name_file=FILE_NAME, integral=True)
    
                # Documentation of the results
                row_data = {
                       "Location": ACT_STATION,
                       "Energy": ACTUAL_ENERGY,
                       "Year": year,
                       "Load Scaling":  input_dict["load-scaling"],
                       "Storage Capacity": st_size,
                       "Storage Time" :  input_dict["storage-hours-max"] ,
    
                   }
                row_data.update( storage_states)
                row_data.update(level_of_autonomy)
                row_data.update( charging)
    
                   # Fügen Sie die Zeilendaten der Gesamt-Datenliste hinzu
                eval_docu.append(row_data)
        df_docu = pd.DataFrame(eval_docu)
    
        # Verwenden Sie 'Location' als Index
        df_docu.set_index("Location", inplace=True)
    
        print(df_docu)





"""

Generate simple energy load curves for analysis of weather data

Either as step function or as sine, differences for weekdays and weekends
included.

B. Atakan, Uni Duisburg-Essen, Germany

SPP2403

2025-01-20

"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

def generate_seasonal_modulation(time_index, amplitude=0.05):
    """
    Generate a seasonal modulation function using a cosine wave.
    
    Parameters
    ----------
    time_index : pandas DateTimeIndex
        The period for which values are generated.
    amplitude : float, optional
        The amplitude of the seasonal variation. Default is 0.05.
    
    Returns
    -------
    pandas Series
        Seasonal modulation factor for each timestamp.
    """
    # Umwandlung der Zeitstempel in den Tagesindex innerhalb eines Jahres (0 bis 364 oder 365)
    day_of_year = time_index.day_of_year
    
    # Kosinusfunktion für die jahreszeitliche Modulation
    seasonal_factor = 1 + amplitude * np.cos(2 * np.pi * day_of_year / 365)

    
    return pd.Series(seasonal_factor, index=time_index)

def generate_load_curve(time_index, step=True, load_parameters=None):
    """
    Generate simple load curves either step function or sine,
    with an additional seasonal modulation.

    Parameters
    ----------
    time_index : pandas DateTimeIndex
        The period for which values are generated.
    step : Boolean, optional
        If True, it is a step function, else a sine. Default is True.
    load_parameters : dictionary, optional
        Values for min/max and weekend/workdays and intermediate values.

    Returns
    -------
    pandas dataframe
        With the load values.
    """
    if load_parameters is None:
        load_parameters = {
            'min_weekday': 0.650,
            'max_weekday': 1.0,
            'min_weekend': 0.650,
            'max_weekend': 0.85,
            'intermediate_fraction': 0.80,
            'multiplier': 1.0,
            'offset': 0.0,
            'normalize': True,
            'amplitude_season':0.05,  # seasonal
        }

    load_curve = pd.DataFrame(index=time_index)
    load_curve['load'] = 0.0

    def generate_sinusoidal_load(t, wert_min, wert_max):
        amplitude = (wert_max - wert_min) / 2
        offset = wert_min + amplitude
        phase_shift = 6
        return offset + amplitude * np.sin((2 * np.pi / 24) * (t - phase_shift))

    for timestamp in load_curve.index:
        hour = timestamp.hour
        is_weekend = timestamp.weekday() >= 5

        if is_weekend:
            if step:
                if hour < 6:
                    load_curve.at[timestamp, 'load'] = load_parameters['min_weekend']
                elif hour < 7:
                    load_curve.at[timestamp, 'load'] = load_parameters['intermediate_fraction']
                elif hour < 22:
                    load_curve.at[timestamp, 'load'] = load_parameters['max_weekend']
                elif hour < 23:
                    load_curve.at[timestamp, 'load'] = load_parameters['intermediate_fraction']
                else:
                    load_curve.at[timestamp, 'load'] = load_parameters['min_weekend']
            else:
                load_curve.at[timestamp, 'load'] = generate_sinusoidal_load(
                    hour, load_parameters['min_weekend'], load_parameters['max_weekend']
                )
        else:
            if step:
                if hour < 6:
                    load_curve.at[timestamp, 'load'] = load_parameters['min_weekday']
                elif hour < 7:
                    load_curve.at[timestamp, 'load'] = load_parameters['intermediate_fraction']
                elif hour < 19:
                    load_curve.at[timestamp, 'load'] = load_parameters['max_weekday']
                elif hour < 22:
                    load_curve.at[timestamp, 'load'] = load_parameters['intermediate_fraction']
                else:
                    load_curve.at[timestamp, 'load'] = load_parameters['min_weekday']
            else:
                load_curve.at[timestamp, 'load'] = generate_sinusoidal_load(
                    hour, load_parameters['min_weekday'], load_parameters['max_weekday']
                )

    # Offset und Skalierung
    load_curve['load'] -= load_parameters['offset']
    load_curve['load'] *= load_parameters['multiplier']

    # Normierung (falls gewünscht)
    l_mean = load_curve['load'].mean()
    if load_parameters['normalize']:
        load_curve["load not normalized"] = load_curve['load']
        load_curve['load'] /= l_mean
        if load_parameters['multiplier']!=1.0:
            print('WARNING, load-multiplier neglected, due to normalization!')

    # Saisonale Modulation
    load_curve['seasonal_factor'] = generate_seasonal_modulation(time_index, amplitude=load_parameters["amplitude_season"])
    load_curve['load'] *= load_curve['seasonal_factor']
    load_curve['load'] = load_curve['load'] /load_curve['load'].mean()

    return load_curve

def _generate_sinusoidal_load(t, wert_min, wert_max):
    """
    Generiert einen sinusförmigen Verlauf mit einer Periode von 24 Stunden,
    einer Phasenverschiebung von 6 Stunden, sowie einem bestimmten
    minimalen und maximalen Wert.

    :param t: Array von Zeitpunkten (in Stunden), von 0 bis 24.
    :param wert_min: Minimalwert der Lastkurve.
    :param wert_max: Maximalwert der Lastkurve.
    :return: Sinusförmige Lastkurve.
    """
    # Definiere die Amplitude und den Offset
    amplitude = (wert_max - wert_min) / 2
    offset = wert_min + amplitude

    # Phasenverschiebung von 6 Stunden für die Sinusfunktion
    phase_shift = 6

    # Generiere die sinusförmige Lastkurve
    load_curve = offset + amplitude * \
        np.sin((2 * np.pi / 24) * (t - phase_shift))

    return load_curve


if __name__ == '__main__':
    time_ind = pd.date_range(start='2021-01-01', end='2022-12-31', freq="h")
    # Beispielaufruf der Funktion
    load_params = {
        'min_weekday': 0.650,
        'max_weekday': 1.0,
        'min_weekend': 0.650,
        'max_weekend': 0.85,
        'intermediate_fraction': 0.80,
        'multiplier': 50.0,
        'offset':0.0,
        'normalize': True,
        'amplitude_season':0.05,  # seasonal
    }

    # For step function
    load_curve_step = generate_load_curve(time_ind,
                                          step=True, load_parameters=load_params)

    # For sinusoidal function
    load_curve_sine = generate_load_curve(time_ind,
                                          step=False, load_parameters=load_params)

    # Plot der ersten zwei Wochen
    N_H = 24*14
    plt.figure(figsize=(12, 6))
    plt.plot(load_curve_step.index[:N_H], load_curve_step['load'][:N_H],
             label='Stufenfunktion', color='blue')  # N_H Stunden = 14 Tage
    plt.plot(load_curve_sine.index[:N_H], load_curve_sine['load']
             [:N_H], label='Sinusfunktion', color='orange')
    plt.title('Lastkurve: Stufenfunktion vs. Sinusfunktion (Erste zwei Wochen)')
    plt.xlabel('Datum')
    plt.ylabel('Last (Skaliert)')
    plt.grid()
    plt.legend()
    plt.show()

    int_step = load_curve_step["load"].sum()
    int_sine = load_curve_sine["load"].sum()

    print(f"Sum/Integral, sine: {int_sine}, setep: {int_step}")
    print("step:\n", load_curve_step.describe())

    print("sine:\n", load_curve_sine.describe())



# -*- coding: utf-8 -*-
"""
Created on Sat Feb  1 11:22:18 2025

@author: atakan
"""
import os
from pathlib import Path

# Standardverzeichnisse
DEFAULT_RESULTS_DIR = os.path.join(os.getenv('TEMP', '/tmp'), 'results', 'storage_residual')
DEFAULT_WEATHER_DATA_DIR = os.path.join(os.getenv('TEMP', '/tmp'), 'data')

# Benutzerdefinierte Verzeichnisse
RESULTS_DIR = os.getenv('RESULTS_DIR', DEFAULT_RESULTS_DIR)
RESULTS_DIR =Path(RESULTS_DIR)

DATA_DIR = Path('../data')
WEATHER_DATA_DIR = DEFAULT_WEATHER_DATA_DIR

# Sicherstellen, dass die Verzeichnisse existieren
os.makedirs(RESULTS_DIR, exist_ok=True)
os.makedirs(DATA_DIR, exist_ok=True)
os.makedirs(WEATHER_DATA_DIR, exist_ok=True)




# -*- coding: utf-8 -*-
"""
Created on Thu Jan 16 11:16:16 2025

@author: atakan
"""

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import requests
import zipfile
import numpy as np
import os

from numpy.polynomial import Polynomial
from scipy import signal

from rws_py import weather_load_storage
from rws_py import generate_load_curve
from rws_py import analyze_residual
from rws_py import RESULTS_DIR, WEATHER_DATA_DIR


pd.set_option('display.max_columns', None)

WITH_LOAD = False  # analyze the weather data togeher with an estimated load or not

SOLAR_THRESHOLD = 0  # J/cm2 [x10000/600 -> W/m²] ; ca. 100 W/m2
# km/h (converted: 1 m/s = 3.6 km/h), energy conversion:cubic
WIND_THRESHOLD = 0


# URLs for DWD data Düsseldorf/Cuxhaven
stations = {"Cuxhaven": '00891',
            "Düsseldorf": '01078',
            'Freiburg': '01443'}
# stations = {"Cuxhaven": '00891'}
# ACT_STATION = "Düsseldorf"
SOLAR_URL_BASE = "https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/solar/historical"
WIND_URL_BASE = "https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/wind/historical"

# SOLAR_ACT = '10minutenwerte_SOLAR_' + \
#     stations[ACT_STATION]+'_20200101_20231231_hist.zip'
# WIND_ACT = '10minutenwerte_wind_' + \
#     stations[ACT_STATION]+'_20200101_20231231_hist.zip'
solar_columns = ['MESS_DATUM', 'QN', 'GS_10']
wind_columns = ['MESS_DATUM', 'QN', 'FF_10']

DATA_SOURCE =  WEATHER_DATA_DIR


def load_data(bse_url_in, filename_in, exp_col, directory=DATA_SOURCE,
              energy_column=2,
              wind_bounds=(3.3, 11.0, 25),
              normalize=True,
              resample=None):
    """
    Load weatherdata from a directory if it already exists,
    and only download and save if it does not exist.


    Parameters
    ----------
    bse_url_in : string
        url from websiteN.
    filename_in : string
        DESCRIPTION.
    exp_col : List
        Which column (names) are expected at least in the file.
    directory : string, optional
        folder to look for data first and store them after downloading, when
        they do not exist. The default is DATA_SOURCE.
    energy_column : Integer, optional
        position of thecolumn within expected_columns, where the solar energy
        density or wind velocity is to be found. The default is 2.
    wind_bounds : list or tuple of length 3, optional
        Minimum, nominal and maximum wind speed of turbine in m/s . The default
        is (3.3, 11.0, 25).
    normalize : Boolean, optional
        for normalizing the "power" column by its mean value, The default
        value is True.
    resample : String/None, optional
        if 'h' the data will be resampled to hourly, taken the mean(other
        resampling would also be ok, check the panadas resample manual). If it
        is None, no resampling will be performed. The default is None.

    Raises
    ------
    IOError
        DESCRIPTION.
    KeyError
        DESCRIPTION.

    Returns
    -------
    df : Pandas DataFrame
        The data with the important values in the power column and the datime
        as index.

    """

    verbose = False
    filepath = os.path.join(directory, filename_in)
    if os.path.exists(filepath):
        print(f"Data found: {filename_in}")
    else:
        url_act = bse_url_in+"/"+filename_in
        print(f"Data not found: {filename_in}. Downloading...")
        response = requests.get(url_act)
        if response.status_code == 200:
            with open(filepath, 'wb') as f:
                f.write(response.content)
            print(f"File downloaded and saved: {filename_in}")
        else:
            raise IOError(f"Failed to download file: {filename_in}")

    with zipfile.ZipFile(filepath) as z:
        csv_filename = z.namelist()[0]
        with z.open(csv_filename) as csv_file:
            df = pd.read_csv(csv_file, sep=None, engine='python', header=0)

    # Debug: Zeige die tatsächlichen Spaltennamen
    if verbose:
        print("Actual columns:", df.columns.tolist())

    df.columns = df.columns.str.strip()
    df = df.replace(-999, 0)

    # Debug: Zeige die Spaltennamen nach strip()
    if verbose:
        print("Columns after strip:", df.columns.tolist())

    # Check for missing columns before proceeding
    missing_columns = [
        col for col in exp_col if col not in df.columns]
    if missing_columns:
        raise KeyError(
            f"The following columns are missing in the data: {missing_columns}")
    energy_column = exp_col[energy_column]

    # Prüfe, ob das Format der MESS_DATUM-Spalte numerisch ist (Wind-Daten)
    if df['MESS_DATUM'].dtype == 'int64' or df['MESS_DATUM'].dtype == 'object' \
        and df['MESS_DATUM'].str.isnumeric().all():
        df['MESS_DATUM'] = pd.to_datetime(
            df['MESS_DATUM'], format='%Y%m%d%H%M')

    else:
        df['MESS_DATUM'] = pd.to_datetime(df['MESS_DATUM'])

    if filename_in.find("SOLAR") > 0:
        df["power"] = df[energy_column]*16.66  # from J/(cm2 h) to W/m2

        df["solar intensity / (J/cm2/10 min)"] = df[energy_column]
    elif filename_in.find("wind") > 0:
        df["power"] = (df[energy_column])**3
        df["velocity / (m/s)"] = df[energy_column]  # km/h ->m/s

        df.loc[(df[energy_column] > wind_bounds[1]),
               "power"] = wind_bounds[1]**3  # reaching the nominal wind speed
        df.loc[(df[energy_column] < wind_bounds[0]) | (df[energy_column]
                                                       > wind_bounds[2]),
               "power"] = 0  # too low or too high velocity

    else:
        print("Problem: Neither wind nor solar data!")


    df.set_index('MESS_DATUM', inplace=True)  # dateime is Index now
    if resample is not None:
        df = df.drop('eor', axis=1).resample(resample).mean()
    df['time_diff_seconds'] = df.index.to_series().diff().dt.total_seconds()

    if normalize:
        # Jahrweise Mittelwert berechnen
        mean_act = df.groupby(df.index.year)['power'].transform('mean')

        # Originalwerte speichern
        df["power not normalized"] = df["power"].copy()

        # Jahrweise Normierung
        df["power"] = df["power"] / mean_act


    return df


def resample_to_hourly(data_in, energy):
    """
    Resample 10-minute data to hourly data by taking the mean.
    """

    data_hourly_in = None
    if energy == "Solar":

        data_hourly_in = data_in.resample('h').mean()

    elif energy == "Wind":

        data_hourly_in = data_in.resample('h').mean()  # BA must be checked!
    return data_hourly_in


def analyze_weather(data, column, threshold, time_per_point=1):
    """
    Analyze weather data based on periods above and below a threshold.
    """
    data = data.copy()
    # Create sign series based on the threshold
    data['above_threshold'] = data[column] > threshold
    data['block'] = (data['above_threshold'].diff() != 0).cumsum()

    # Calculate period length in hours (hourly intervals -> hours)
    data['count'] = 1  # 1 hour per row, for aggregation
    data['count'] = data.groupby('block')['count'].transform(
        'sum') * time_per_point  # Convert to hours

    # Calculate the total number of inactive periods per block (scaled to hours)
    df_period = data.groupby('block')[column].agg(
        ['count', 'mean', 'std', 'sum', 'min', 'max']).reset_index()

    # Calculate the mean for active periods
    df_period['mean'] = df_period.apply(
        lambda row: data[data['block'] == row['block']][column].mean(), axis=1)
    df_period['count'] = df_period['count'] * time_per_point

    # Lists for active-inactive and inactive-active combinations
    active_inactive_list = []
    inactive_active_list = []

    for i in range(len(df_period)):
        current_row = df_period.iloc[i]
        current_row = current_row.copy()
        if i + 1 < len(df_period):
            next_row = df_period.iloc[i + 1]
            if current_row['mean'] > threshold:
                # Active-Inactive combination
                if not df_period.iloc[i + 1]['mean'] > threshold:
                    count_inv = next_row['count']
                    # Add the value as a new column in current_row
                    current_row['count_inv'] = count_inv
                    active_inactive_list.append(current_row)
            else:
                # Inactive-Active combination
                if df_period.iloc[i + 1]['mean'] > threshold:
                    count_inv = next_row['count']
                    # Add the value as a new column in current_row
                    current_row['count_inv'] = count_inv
                    inactive_active_list.append(current_row)

    # Ensure that the lists for active and inactive periods are 2D data
    # Use `.concat()` on the list of DataFrames and ensure the DataFrame is created correctly
    df_active_inactive = pd.DataFrame(active_inactive_list)
    df_inactive_active = pd.DataFrame(inactive_active_list)

    return df_period, df_active_inactive, df_inactive_active


# Modifizierte Visualisierungsfunktion


def plot_enhanced_analysis(data, results, column, threshold, en_t, year, fil_name):
    """
    Create enhanced visualizations for the analysis results with improved period distribution plots
    """
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(15, 12))

    # Plot 1: Time series with thresholds
    part = data.iloc[7000:10000]
    sns.lineplot(data=part, y=column, x="MESS_DATUM", ax=ax1)
    ax1.axhline(threshold, color='k', label='Basic threshold', linestyle='--')
    ax1.axhline(results['direct_threshold'], color='r',
                label='Direct usage threshold', linestyle=':')
    ax1.set_title(f'Time Series with Thresholds\n{en_t}, {year}')
    ax1.legend()

    # Plot 2: Period Length Distributions (log scale)
    # Verbesserte Darstellung der Periodenlängen mit logarithmischer Skala
    ax2.set_xscale('linear')
    ax2.set_yscale('linear')
    bins = np.linspace(
        1, max(results['active_inactive']['count'].max(), 1), 30)
    if not results['active_inactive'].empty:
        sns.histplot(data=results['active_inactive'], x='count', ax=ax2,
                     label='Basic threshold periods', bins=bins, alpha=0.6)

    if not results['direct_period_data'].empty:
        sns.histplot(data=results['direct_period_data'], x='count', ax=ax2,
                     label='Direct usage periods', color='r', bins=bins, alpha=0.4)

    ax2.set_title('Period Length Distribution')
    ax2.set_xlabel('Period length (hours)')
    ax2.set_ylabel('Count')
    ax2.legend()

    # Plot 3: Scatter plot of period lengths vs intensities
    if not results['active_inactive'].empty:
        sns.scatterplot(data=results['active_inactive'],
                        x='count', y='mean',
                        size='sum',
                        alpha=0.6,
                        ax=ax3)
        ax3.set_xscale('linear')
        ax3.set_title('Period Length vs Intensity')
        ax3.set_xlabel('Period length (hours)')
        ax3.set_ylabel('Mean intensity')

    # Plot 4: Cumulative distribution of period lengths
    if not results['active_inactive'].empty:
        counts_sorted = np.sort(results['active_inactive']['count'])
        cumulative = np.arange(1, len(counts_sorted) + 1) / len(counts_sorted)
        ax4.plot(counts_sorted, cumulative, 'b-', label='Basic threshold')

    if not results['direct_period_data'].empty:
        counts_sorted_direct = np.sort(results['direct_period_data']['count'])
        cumulative_direct = np.arange(
            1, len(counts_sorted_direct) + 1) / len(counts_sorted_direct)
        ax4.plot(counts_sorted_direct, cumulative_direct,
                 'r-', label='Direct usage')

    ax4.set_xscale('log')
    ax4.set_title('Cumulative Distribution of Period Lengths')
    ax4.set_xlabel('Period length (hours)')
    ax4.set_ylabel('Cumulative probability')
    ax4.legend()

    plt.tight_layout()
    plt.savefig(fil_name + "-enhanced-analysis.png",
                bbox_inches='tight', dpi=300)
    plt.show()


def plot_analysis(data, results, column, threshold, en_t, year, fil_name,
                  period_pause,
                  bins=[0, 1, 2, 4, 8, 16, 32, 64, 128, 200]):
    """
    Create  visualizations for the analysis results with improved period distribution plots
    """
    # Visualize results
    fig0 = plt.figure()

    sns.scatterplot(data=results['direct_inactive'],
                    x='count', y='count_inv', size='sum', hue='mean')
    plt.title(f'Active vs Inactive Periods, {en_t}, {year}')
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0)
    plt.tight_layout()
    plt.savefig(str(fil_name)+"-periods.png")
    plt.show()
    # plt.figure()
    histo2, xedges, yedges = np.histogram2d(
        results['direct_inactive']['count_inv'], results['direct_inactive']['count'], bins=bins)

    fig0, ax0 = plt.subplots(1)

    ax0.imshow(histo2)
    xranges = [
        f"period {xedges[i]:.1f} - {xedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    yranges = [
        f"pause {yedges[i]:.1f} - {yedges[i+1]:.1f}" for i in range(len(xedges) - 1)]
    ax0.set_xticks(np.arange(len(xranges)), labels=xranges,
                   rotation=45, ha="right", rotation_mode="anchor")
    ax0.set_yticks(np.arange(len(yranges)), labels=yranges)
    for i in range(len(yranges)):
        for j in range(len(xranges)):
            if histo2[i, j] > 0:
                ax0.text(j, i, f"{histo2[i, j]:.0f}",
                                ha="center", va="center", color="w", fontsize=10)

    ax0.set_title(fil_name+", 2D Histogram")
    fig0.tight_layout()
    fig0.savefig(fil_name+"-2D-hist.png")
    plt.show()
    df = pd.DataFrame(histo2.T, index=yranges, columns=xranges)

    # xedges als eigene Spalte hinzufügen
    df.insert(0, "y_edges_start", yedges[:-1])

    # CSV speichern
    df.to_csv(fil_name+"histogram_data.csv")

    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(ncols=2, nrows=2)

    pcm = ax1.pcolormesh(xedges, yedges, histo2, cmap='rainbow')
    fig.colorbar(pcm, ax=ax1)
    ax1.set_xlabel("Period length / h")
    ax1.set_ylabel("pause length / h")
    sum_count_fit = Polynomial.fit(
        results['direct_inactive']['count'], results['direct_inactive']['sum'], 2)
    ax2.plot(results['direct_inactive']['count'],
             results['direct_inactive']['sum'], "+k")
    ax2.plot(results['direct_inactive']['count'], sum_count_fit(
        results['direct_inactive']['count']), '.')
    ax2.set_xlabel("Period length / h")
    ax2.set_ylabel("total intensity")
    ax3.plot(period_pause['count'], period_pause['mean'], "+k")
    ax3.set_xlabel("Period length / h")
    ax3.set_ylabel("mean intensity")

    act_corr = signal.correlate(data[column], data[column])
    # f, Pxx_den = signal.periodogram(data[expected_columns[2]], fs=1/600)
    # ax4.semilogy(f, Pxx_den)
    ax4.plot(act_corr)
    ax4.set_xlabel("time shift / h")
    ax4.set_ylabel("autocorrelation")
    fig.suptitle(en_t)
    fig.savefig(fil_name+"-cor-hist.png")
    plt.show()

    # for illustration
    fff, (aaa, aab) = plt.subplots(2)
    part = data.iloc[7000:10000]
    sns.lineplot(data=part, y=column, x="MESS_DATUM", ax=aaa)
    aaa.axhline(threshold, color='k')
    aaa.set_title(en_t)
    # Erstelle den Inlay-Plot
    # Erstellen Sie Achsen für den Inlay-Plot

    sns.lineplot(data=data, y=column, x="MESS_DATUM", ax=aab)
    aab.axhline(threshold, color="k")
    aab.set_title('All Data')  # optional
    fff.savefig(fil_name+"-timeseries.png")
    plt.show()
    fig_n, axn = plt.subplots(1)

    results['direct_inactive'].hist(
        column='count_inv', ax=axn, bins=bins, log=True)
    results['direct_inactive'].hist(
        column='count', ax=axn, bins=bins, rwidth=.75, log=True)
    axn.set_xlabel("period/pause duration / h")
    axn.legend(["pause", 'with'])

    axn.set_title(en_t+", histogram")


if __name__ == "__main__":
    VERBOSE = True

    columns = ["Station", "Year", "power", "Threshold", "sum", "mean",
               "10 min periods", "periods>1h", "sum with energy",
               "sum without energy", "ratio of period lengths",
               "mean_active_periods", "direct_threshold",
               "storage_periods_count", "period_mean_duration", "mean_pause_duration",
               'std_pause_duration',
               'std_duration']

    evaluation = pd.DataFrame(columns=columns)
    outp = []
    histos = []
    BINS = 30

    all_dat = []
    ENERGY_COLUMN = "power"  # Name of the column with the energy proportional data

    # Prepare data sources
    for ACT_STATION in stations.keys():
        SOLAR_ACT = '10minutenwerte_SOLAR_' + \
            stations[ACT_STATION]+'_20200101_20231231_hist.zip'
        WIND_ACT = '10minutenwerte_wind_' + \
            stations[ACT_STATION]+'_20200101_20231231_hist.zip'
        all_dat.append([SOLAR_URL_BASE, SOLAR_ACT,
                       solar_columns, SOLAR_THRESHOLD, ACT_STATION])
        all_dat.append([WIND_URL_BASE, WIND_ACT, wind_columns,
                       WIND_THRESHOLD, ACT_STATION])

    # Main analysis loop
    for dat_act in all_dat:
        base_url = dat_act[0]
        filename = dat_act[1]
        expected_columns = dat_act[2]
        thresh_out = dat_act[3]
        ACT_STATION = dat_act[4]

        # Load data and  # Resample data to hourly
        data_all = load_data(base_url, filename, expected_columns,
                             resample="h")

        # Process each year
        for year in range(2020, 2022):
            print(f"year: {year}\n")
            en_t = ACT_STATION+", "
            if dat_act[0].count('wind'):
                act_energy = "Wind"
            else:
                act_energy = "Solar"
            en_t += act_energy
            f_name = f"{en_t}, {year}"
            f_base=f_name.replace(", ", "-")
            dir_name = RESULTS_DIR

            fil_name = dir_name / f_base
            data_hourly = data_all[data_all.index.year == year]

           
            description = data_hourly.describe()
            description.to_csv(f"{fil_name.stem}-describe.csv", sep=";")

            if WITH_LOAD:
                load_par = {
                    'min_weekday': 0.650,
                    'max_weekday': 1.0,
                    'min_weekend': 0.650,
                    'max_weekend': 0.85,  # only for step-function
                    'intermediate_fraction': .75,
                    # scales everything
                    'multiplier': data_hourly["power"].mean()
                }
                load_df = generate_load_curve(
                    data_hourly.index, load_parameters=load_par)
                with_load = weather_load_storage(data_hourly, load_df)
                # BA with load must be stored and evaluated/plotted

            if VERBOSE:
                print(description)
            # Perform enhanced analysis (includes basic analysis)
            if WITH_LOAD:
                data_hourly = with_load.copy()
                results_act = analyze_residual(data_hourly, "residual-no-storage")

            else:
                pass

            # Extract results for further processing
            period_pause = results_act['active_inactive']
            long_per_pause = period_pause[period_pause['count'] >= 1]

            # Calculate metrics
            sum_all = data_hourly[ENERGY_COLUMN].sum()
            mean_all = data_hourly[ENERGY_COLUMN].mean()
            sh_periods = period_pause['count'].count()
            long_periods = long_per_pause['count'].count()
            periods_with_energy = period_pause['count'].sum()
            periods_without_energy = period_pause['count_inv'].sum()
            per_ratio = periods_with_energy / \
                periods_without_energy if periods_without_energy != 0 else 0

            # Print analysis results
            print(
                f"Threshold:{thresh_out:.1f}, No.-10 min periods: {sh_periods}, above 1h: {long_periods}")
            print(
                f"Sum of all times, with energy {periods_with_energy:.1f}, without {periods_without_energy:.1f}, ratio:{per_ratio:.3f} \n")

            # Create evaluation row
            eval_row = [
                ACT_STATION,
                year,
                act_energy,
                thresh_out,
                sum_all,
                mean_all,
                sh_periods,
                long_periods,
                periods_with_energy,
                periods_without_energy,
                per_ratio,
                results_act['basic_stats']['mean_active_periods'],
                results_act['direct_threshold'],
                results_act['storage_periods']['count'],
                results_act['storage_periods']['mean_duration'],
                results_act['storage_periods']['mean_pause_duration'],
                results_act['storage_periods']['std_pause_duration'],
                results_act['storage_periods']['std_duration'],
            ]

            # Add to evaluation DataFrame
            evaluation.loc[len(evaluation)] = eval_row

            # Create visualizations
            plot_analysis(
                data_hourly, results_act, ENERGY_COLUMN, thresh_out, en_t, year,
                str(fil_name)+"-H-",
                period_pause)
            plot_enhanced_analysis(
                data_hourly, results_act, ENERGY_COLUMN, thresh_out, en_t, year,
                str(fil_name)+"-H-")

    # Save final results
    evaluation.to_csv(dir_name/("H-"+"all-eval-enhanced.csv"),
                      sep=";", encoding='utf-8')



"""Top-level package for residuals_weather_storage.

rws_py package.

DFG - SPP2403

B.Atakan, Univ. Duisburg-Essen
2025-02-07

"""

import pkgutil
from .config import RESULTS_DIR, DATA_DIR, WEATHER_DATA_DIR
from .analyze_weather_load_storage import weather_load_storage, calc_residuals
from .load_curve_simple import generate_load_curve
from .residual_analyse import analyze_residual, storage, esm_residual_to_df
from .dwd_analyze_store_second_threshold import load_data, analyze_weather, \
    resample_to_hourly, plot_analysis, plot_enhanced_analysis

from .help_funcs import append_suffix_to_path as app_suf
from .help_funcs import get_stations
from .analysis_universal import residual_analyse, plot_analysis_b, plot_res,\
    plot_histo_analysis,plot_histo_int_analysis, posterior_analysis_residual,\
        print_dict


__author__ = """Burak Atakan"""
__email__ = 'atakan.thermodynamik.duisburg@gmail.com'
__version__ = '0.2.1.post1'

__all__ = [
    'weather_load_storage',
    'calc_residuals',
    'generate_load_curve',
    'analyze_residual',
    'residual_analyse',
    'plot_analysis_b',
    'plot_res',
    'plot_histo_analysis',
    'plot_histo_int_analysis',
    'posterior_analysis_residual',
    'print_dict',
    'storage',
    'esm_residual_to_df',
    'RESULTS_DIR',
    'DATA_DIR',
    'WEATHER_DATA_DIR',
    'load_data',
    'analyze_weather',
    'resample_to_hourly',
    'plot_analysis',
    'plot_enhanced_analysis',
    'help_funcs',
    'app_suf',

]

stations = get_stations() # all DWD weather stations for 10 min data->dictionary

__path__ = pkgutil.extend_path(__path__, __name__)

# Automatisches Importieren aller Untermodule (inkl. utils)
for finder, module_name, ispkg in pkgutil.walk_packages(__path__, __name__ + "."):
    __import__(module_name)




# -*- coding: utf-8 -*-
"""
Analyse the rsidual loads from different countries statistically

with respect to the correlation of periods with positive and negative residual loads.
sum: integral (Energy)
mean: measure for Power
count: period (in h)
and further values are also calculated like the 25%, 75% etc.

Context: SPP2403 Cooperation Atakan / Bertsch

Created on Thu Dec  5 10:00:37 2024

@author: atakan

"""

import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sbn
from scipy import signal

from rws_py import DATA_DIR, RESULTS_DIR
plt.close("all")

# Beispiel-Daten (ersetze durch deine eigene Datenserie)
# data_0 = pd.DataFrame(
#     {'Residuallast': [5, 1, 2, -2, -1, -3, 5, 2, -4, -6, 8, 1, 6, 7, 8]})
test = False
dir_name = RESULTS_DIR
f_name_base = dir_name/"res_store"

def analyze_residual(data, column, verbose=False):
    """
    Analyze Residual load data for periods of negative residuals followed by positive

    with respect to time (length), integral(sum)=energy and the mean values
    of each period.

    Parameters
    ----------
    data : pd.DataFrame
        to be analyzed.
    column : string
        the column with the data to be analyzed.
    verbose : Boolean, optional
        if you want some printing. The default is False.

    Raises
    ------
    KeyError
        DESCRIPTION.

    Returns
    -------
    results : dictionary with three entries, as follows
        -""
    df_period : pd.DataFrame
        all periods ith the folowing one of opposite sign .
    df_pos_neg : pd.DataFrame
        positive residual load  periods with the folowing one of opposite sign .
    df_neg_pos : pd.DataFrame
        negative residual load  periods with the folowing one of opposite sign .

    """

    # Vorzeichenserien erstellen
    data['sign'] = np.sign(data[column])

    # Finden Sie die Perioden positiver und negativer Zahlen
    data['block'] = (data['sign'].diff() != 0).cumsum()

    # Statistik für jede Periode erstellen
    df_period = data.groupby('block')[column].agg(
        ['count', 'mean', 'std', 'sum', 'min', 'max']).reset_index()
    df_period['25%'] = data.groupby('block')[column].quantile(0.25).values
    df_period['50%'] = data.groupby('block')[column].quantile(0.5).values
    df_period['75%'] = data.groupby('block')[column].quantile(0.75).values
    df_period['positive'] = df_period['sum'] > 0

    # Listen für Positive-Negative und Negative-Positive Kombinationen
    pos_neg_list = []
    neg_pos_list = []

    # Iteriere durch die Perioden und kombiniere sie
    for i in range(len(df_period)):
        current_row = df_period.iloc[i]

        if current_row['positive']:
            # Positive-Negative Kombination
            if i + 1 < len(df_period) and not df_period.iloc[i + 1]['positive']:
                next_row = df_period.iloc[i + 1]
                combined = pd.concat([current_row.add_suffix(
                    '_pos'), next_row.add_suffix('_neg')])
            else:
                zero_row = pd.Series(
                    0, index=current_row.index).add_suffix('_neg')
                combined = pd.concat(
                    [current_row.add_suffix('_pos'), zero_row])
            pos_neg_list.append(combined)
        else:
            # Negative-Positive Kombination
            if i + 1 < len(df_period) and df_period.iloc[i + 1]['positive']:
                next_row = df_period.iloc[i + 1]
                combined = pd.concat([current_row.add_suffix(
                    '_neg'), next_row.add_suffix('_pos')])
            else:
                zero_row = pd.Series(
                    0, index=current_row.index).add_suffix('_pos')
                combined = pd.concat(
                    [current_row.add_suffix('_neg'), zero_row])
            neg_pos_list.append(combined)

    # DataFrames erstellen
    df_pos_neg = pd.DataFrame(pos_neg_list).reset_index(drop=True)
    df_neg_pos = pd.DataFrame(neg_pos_list).reset_index(drop=True)

    # Ergebnisse anzeigen
    if verbose:
        print("Positive-Negative Perioden (fehlende negative Werte mit Nullen aufgefüllt):")
        print(df_pos_neg)

        print(
            "\nNegative-Positive Perioden (fehlende positive Werte mit Nullen aufgefüllt):")
        print(df_neg_pos)
    return df_period, df_pos_neg, df_neg_pos




def storage(power, actual_capacity, full_capacity, efficiency=1, time_step=1,
            sp_names=None):
    """
    Simple storage model without power restriction but max. capacity.

    Parameters
    ----------
    power : Float
        power entering the storage (=positive) or wanted(=negative), if possible
                    with the actual state of charge.
    actual_capacity : Float
        what is the actual charge (energy).
    full_capacity : Float
        what is the full capacity.
    efficiency : Float, optional
        discharging efficiency. The default is 1.
    time_step : Float, optional
        time step relative to the next value in hours. The default is 1.
    sp_names : list of strings, optional
        the names of the three values calculated. The default is None.
    Returns
    -------
    result_dict : Dictionary

        - "act_st_cap" actual loading/energy after the time step
        - "power_st_in": the real power entering/exiting the storage
        - "SOC": state of charge.

    """

    if power >= 0:  # Laden
        if actual_capacity < full_capacity - power*time_step:
            pass
        elif actual_capacity < full_capacity:
            power = (full_capacity - actual_capacity)/time_step
        else:
            power = 0
        actual_capacity += time_step*power
    else:  # Entladen
        power_with_efficiency = power / efficiency  # Berücksichtige Effizienz bei der Entladeleistung
        if actual_capacity > np.abs(time_step*power_with_efficiency):
            pass
        elif actual_capacity > 0:
            power = -actual_capacity*efficiency/time_step  # Anpassen der Leistung basierend auf verfügbarer Kapazität
        else:
            power = 0
        actual_capacity += time_step * power_with_efficiency

    # Sicherheitscheck für numerische Ungenauigkeiten
    actual_capacity = max(0, min(actual_capacity, full_capacity))

    soc = actual_capacity/full_capacity
    if not sp_names:
        sp_names = ["act_st_cap", "power_st_in", "SOC"]
    result_dict = dict(zip(sp_names, [actual_capacity, power, soc]))

    return result_dict

def create_filenames(value, base=None):

    exponent = int(np.floor(np.log10(value)))  # Berechne den Exponenten
    base_value = value / (10 ** exponent)  # Basiswert berechnen

    # Runden die Basis auf 3 signifikante Stellen und ohne Dezimalstellen
    # Multiplizieren und umwandeln in Ganze Zahl zur Eliminierung der Dezimalstellen
    rounded_value = round(base_value, 3)

    # Entferne Punkte durch Umwandlung in integer und formatiere den Namen
    filename = f"{int(rounded_value * 1000)}E{exponent - 3}"

    return base+filename




def eval_df_pos_neg(df, col, time=1, verbose=False):
    positive_residuals = df[df[col] > 0][col]
    positive_mean = positive_residuals.mean()
    positive_sum = positive_residuals.sum()
    positive_count = positive_residuals.count()
    # Berechnung für negative Residualwerte
    negative_residuals = df[df[col] < 0][col]
    negative_mean = negative_residuals.mean()
    negative_sum = negative_residuals.sum()
    negative_count = negative_residuals.count()
    if verbose:
        # Ausgabe der Ergebnisse
        print(f"""Positive Residuals: Mittelwert: {
              positive_mean:.2e}, Summe: {positive_sum:.2e}""")
        print(f"""Negative Residuals: Mittelwert: {
              negative_mean:.2e}, Summe: {negative_sum:.2e}""")
        print(f"""summe Residuals: Mittelwert: {
              positive_mean+negative_mean:.2e},
            Summe: {positive_sum+negative_sum:.2e}""")

    return np.array([positive_mean, positive_sum, positive_count, negative_mean,
                     negative_sum, negative_count])

def calc_capacity_dependence(in_ser, m_cap, EFF_STOR=0.75,
                             make_df=False,
                             verbose = False,
                             file_name = 'cap_res-',
                             title = None):
    """
    Calculate a storage (dis-)charging as a function of residual loads

    Parameters
    ----------
    in_ser : pd.Series
        hourly Residual loads (neg->entering).
    m_cap : array of floats
        storage capacities for which the behaviour is calculated.
    EFF_STOR : Float, optional
        discharging efficiency of the storage. The default is 0.75.
    make_df : Boolean, optional
        shall all results be returned as pd.DataFrame or as some numpy.arrays.
        The default is False.
    verbose : Boolean, optional
        if printing is wanted. The default is False.
    file_name : string, optional
        place to store the results, the name will be extended. The default
        is 'cap_res-'.

    Returns
    -------
    Dataframe
        assessment as  function of the residual loads time series.
    or numpy arrays
        if make_df=False, cap_dep_res, st_state, la, cap_change.

    """
    cap_dep_res = []
    st_state = []
    cap_change = []
    la = []
    in_df = pd.DataFrame(in_ser)
    in_df.columns = ["residual"]

    for mca in m_cap:
        cap = 0  # Initiale Speicherkapazität

        fn_act = create_filenames(mca, file_name)

        results = []
        for po in in_df["residual"]:
            st_res = storage(-po, cap, mca, efficiency=EFF_STOR)
            results.append(st_res)
            cap = st_res["act_st_cap"]
        results = pd.DataFrame(results)
        #results.set_index('row_num', inplace=True)
        in_df.set_index(results.index, inplace=True)


        actual_df = pd.concat([in_df, results], axis=1)
        actual_df["row_num"] = actual_df.index
        actual_df["pow_rest"] = actual_df["residual"] + actual_df["power_st_in"]

        # Berechne die Differenzen und weitere Statistiken
        actual_df['change_act_st_cap'] = actual_df['act_st_cap'].diff()

        fi_, ax_ = plt.subplots(3)
        sbn.lineplot(data=actual_df, x="row_num", y="SOC",
                        ax=ax_[0], color='b')
        sbn.lineplot(data=actual_df, x="row_num", y="pow_rest",
                        ax=ax_[1], color='b')
        sbn.lineplot(data=actual_df, x="row_num", y="residual",
                        ax=ax_[2], color='k')
        ax_[0].set_title(title+f", Cap: {mca:2.2e}")
        fi_.savefig(fn_act+".jpg")



        soc_counts = actual_df['SOC'].value_counts()
        zero_count = soc_counts.get(0, 0)  # Anzahl der 0en
        one_count = soc_counts.get(1, 0)    # Anzahl der 1en



        # Berechnung der Differenzen zwischen aufeinanderfolgenden Werten
        actual_df['previous_SOC'] = actual_df['SOC'].shift(
            1)  # Vorheriger Wert

        # Bedingung für Wechsel von 0 auf einen höheren Wert
        from_0_to_higher = ((actual_df['previous_SOC'] == 0) & (
            actual_df['SOC'] > 0)).sum()

        # Bedingung für Wechsel von 1 auf einen niedrigeren Wert
        from_1_to_lower = ((actual_df['previous_SOC'] == 1) & (
            actual_df['SOC'] < 1)).sum()

        # Ausgabe der Ergebnisse
        if verbose:
            print(f"Max. storage capacity: {mca}")
            print(f"""Anzahl der Werte 0 in 'SOC': {
                  zero_count} und 1 in 'SOC': {one_count}""")
            print(f"""Anzahl der Wechsel von 0 auf einen höheren Wert: {
                  from_0_to_higher}; von 1 auf einen niedrigeren Wert: {
                from_1_to_lower}""")

        st_state.append(np.array([zero_count,
                                 one_count,
                                 from_0_to_higher,
                                 from_1_to_lower]))


        # Zugrunde liegende Werte zur LA Auswertung
        time_pos_res_remain = actual_df[actual_df['pow_rest'] > 0].count()['pow_rest']
        time_neg_res_remain = actual_df[actual_df['pow_rest'] < 0].count()['pow_rest']
        time_no_res_remain = actual_df[actual_df['pow_rest'] == 0].count()['pow_rest']

        la_pos = 1 - (time_pos_res_remain / actual_df["pow_rest"].count())
        la_neg = 1 - (time_neg_res_remain / actual_df["pow_rest"].count())
        la_eq = (time_no_res_remain / actual_df["pow_rest"].count())
        la.append(np.array([la_pos, la_neg, la_eq]))

        cap_change.append(eval_df_pos_neg(actual_df, "change_act_st_cap"))

        # Ergebnisse sammeln
        pos_neg_stat = eval_df_pos_neg(actual_df, "pow_rest")
        cap_dep_res.append(np.array([*pos_neg_stat, actual_df["act_st_cap"].iloc[-1]]))
        actual_df.to_csv(fn_act+".csv", sep=";")



    cap_dep_res = np.array(cap_dep_res)
    st_state = np.array(st_state)
    la = np.array(la)
    cap_change= np.array(cap_change)

    # Rückgabe je nach Bedarf



    if make_df:
        cd_names=['load_pos_mean', 'load_pos_sum', 'load_pos_count', 'load_neg_mean', 'load_neg_sum', 'load_neg_count', "final_storedCharge"]
        st_names =["empty", "full", "0->up", "1->down"]
        la_names= ["LA+", "LA-", "eq_ratio"]
        cap_change_names=['load_diff_pos_mean', 'load_diff_pos_sum', 'load_diff_pos_count', 'load_diff_neg_mean', 'load_diff_neg_sum', 'load_diff_neg_count']
        col_names = [ *cd_names, *st_names, *la_names, *cap_change_names]
        combined_df = pd.DataFrame(np.hstack((cap_dep_res, st_state, la, cap_change)), columns=col_names)
        combined_df["storage capacity"] = m_cap
        combined_df["Eff_dis"] = EFF_STOR
        combined_df.to_csv(file_name+"_cap_dependence.csv",sep=";")
        return combined_df
    else:
        return cap_dep_res, st_state, la, cap_change


def esm_residual_to_df(country=None, file_in=None,
                       col_name="residual-no-storage", first_day=None,
                       save_csv=True, verbose=False,
                       get_countries=False,
                       normalize=True):
    """
    Reads ESM residual load data from a CSV or Excel file, extracts the data for a specific country,
    sets the date as the index, and optionally normalizes the data.

    Parameters:
    -----------
    country : str, optional
        The country for which the residual load data should be extracted.
        Defaults to "DE0 Tot" if not specified.

    file_in : str, optional
        The filename (without extension) containing the residual load data.
        If None, defaults to '2024_Bertsch_residual_Load'.

    col_name : str, optional
        The name of the column to store the extracted residual load data.
        Default is "residual-no-storage".

    first_day : str, optional
        Not used in the current function.

    save_csv : bool, optional
        If True, saves the extracted and transposed data to a CSV file for future use. Default is True.

    verbose : bool, optional
        If True, prints the first few rows of the dataset and the index for debugging purposes. Default is False.

    get_countries : bool, optional
        If True, returns a list of available country names from the dataset instead of extracting data. Default is False.

    normalize : bool, optional
        If True, normalizes the extracted residual load data by computing
        the absolute mean of the negative values (=> positive!).
        The original values are then divided by this value. Default is True.

    Returns:
    --------
    pd.DataFrame or list
        - If `get_countries` is True: Returns a list of available country names.
        - Otherwise: Returns a DataFrame with extracted residual load data for the specified country.

    Raises:
    -------
    ValueError
        If the specified file cannot be loaded.

    FileNotFoundError
        If neither the CSV nor the Excel file is found.

    Notes:
    ------
    - The function first attempts to read data from a CSV file. If the CSV is missing, it tries to read from an Excel file.
    - If an Excel file is used, the dataset is transposed, and a datetime index is generated starting from '2050-01-01 00:00:00' with hourly frequency.
    - Normalization is performed only if the normalization factor (difference between mean positive and mean negative values) is nonzero.
    """

    COL_N_ORIGINAL = "residual-original"
    if file_in is None:
        file_in = '2024_Bertsch_residual_Load'

    # **Pfad zum data-Ordner setzen**
    base_dir = os.path.dirname(os.path.abspath(__file__))  # Ordner von residual_analyse.py
    data_dir = os.path.join(base_dir, "..", "data")  # Wechselt eine Ebene hoch und geht in "data"
    file_csv = os.path.join(data_dir, file_in + ".csv")
    file_xlsx = os.path.join(data_dir, file_in + ".xlsx")

    df_transponiert = None  # Initialisierung

    try:
        df_transponiert = pd.read_csv(file_csv, index_col=0)
    except:
        try:
            df = pd.read_excel(file_xlsx)
            df_transponiert = df.set_index('node').T
            start_date = '2050-01-01 00:00:00'
            date_range = pd.date_range(start=start_date, periods=8760, freq='h')
            df_transponiert.index = date_range
            if save_csv:
                df_transponiert.to_csv(file_csv)
        except FileNotFoundError:
            print(f"Fehler: Datei '{file_csv}' oder '{file_xlsx}' nicht gefunden.")
        except Exception as e:
            print(f"Ein unerwarteter Fehler ist aufgetreten: {e}")

    if df_transponiert is None:
        raise ValueError(f"Konnte Datei '{file_csv}' oder '{file_xlsx}' nicht laden!")

    if verbose:
        print(df_transponiert.head())
        print(df_transponiert.index)

    if get_countries:
        return list(df_transponiert.columns)

    if country is None:
        country = "DE0 Tot"

    actual_values = df_transponiert[country].values
    actual_country_df = pd.DataFrame(actual_values, columns=[col_name])
    actual_country_df[COL_N_ORIGINAL] = actual_country_df[col_name]
    actual_country_df.index = pd.to_datetime(df_transponiert.index)
    if normalize:
        actual_values = actual_country_df[COL_N_ORIGINAL].values
        hours = (actual_country_df.index[-1] - actual_country_df.index[0]).total_seconds() / 3600

        # Compute mean of the negative values separately
        # this is, what could be stored on average per year
        norm_factor = abs(actual_values[actual_values < 0].sum()) \
            / hours

        
        
        actual_country_df.loc[:, 'residual-not-normalized'] = actual_country_df[col_name]

        if norm_factor != 0:  # Avoid division by zero
            actual_country_df[col_name] = actual_country_df[col_name]/ norm_factor
        else:
            pass  # Keep original if no variation





    return actual_country_df


if __name__ == "__main__":
        
    test_countries = esm_residual_to_df(get_countries =True)
    test_df = esm_residual_to_df()
    
    EFF_STOR = .75
    m_cap = np.linspace(1e6, 20e6, 15)
    
    if test:
        verbose = False
        df = pd.read_excel('2024_Bertsch_residual_Load.xlsx')
    
        # Lade die Excel-Datei mit pandas
    
        # output_folder = create_output_folder()
        # Transponiere die Tabelle
        df_transponiert = df.set_index('node').T
        df_transponiert['time_diff_seconds'] = np.zeros(
            len(df_transponiert)) + 3600
    
        df_transponiert['time'] = np.arange(0, len(df_transponiert) * 3600., 3600)
        countries = ["DE0 Tot", "IT0 0", "GB5 0", "ES0 0", "IE4 0"]
        spalten = ["act_st_cap", "power_st_in", "SOC"]
        actual_df = df_transponiert["DE0 Tot"].values
        actual_df = pd.DataFrame(actual_df, columns=["residual"])
    
    
        # Beispielaufruf der neuen Funktion
    
        results_df = calc_capacity_dependence(actual_df,
                                              m_cap,
                                              EFF_STOR,
                                              make_df=True,
                                              file_name=f_name_base)
    
        # Wenn du die NumPy-Arrays willtest
        cap_dep_res, st_state, la, cap_change = calc_capacity_dependence(actual_df,
                                                                         m_cap,
                                                                         EFF_STOR,
                                                                         make_df=False,
                                                                         file_name=f_name_base)
    
        fi_n, (ax_n, ax_s)= plt.subplots(2)
        fi_n2, (ax_c, ax_e)= plt.subplots(2)
    
        # Energy integral
        ax_n.plot(m_cap, cap_dep_res[:, 1], "k", label="pos res")
        # integral of remaining negative residuals
        ax_n.plot(m_cap, -cap_dep_res[:, 4], "b:", label="neg res")
        ax_n.plot(m_cap, cap_dep_res[:, -1], "bv",
                  label = "stored finally")  # am Ende im Speicher
        ax_n.plot(m_cap, cap_dep_res[:, 1] + \
                  cap_dep_res[:, 4], "kv-", label="sum res")
        ax_n.plot(m_cap, cap_dep_res[:, 1]+cap_dep_res[:,
                  4]+cap_dep_res[:, -1], "ro", label="sum")
        ax_n.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_n.set_xlabel("storage capacity")
        ax_n.set_ylabel("integral energy")
    
    
        # Storage loads
        ax_s.plot(m_cap, st_state[:, 0], "k", label="empty")
        ax_s.plot(m_cap, st_state[:, 1]/20, "b", label="full/20")
        ax_s.plot(m_cap, st_state[:, 2], "k", label="0->up")
        ax_s.plot(m_cap, st_state[:, 3], "k", label="1->down")
        ax_s.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_s.set_xlabel("storage capacity")
        ax_s.set_ylabel("number of states")
        # Raum für die Legende schaffen
        fi_n.tight_layout(rect=[0, 0, 0.85, 1])
        fi_n.savefig(f_name_base+"-number_cap-plot.png")
    
        # Storage capacity change
        ax_c.plot(m_cap, cap_change[:, 1], "k", label="charging")
        ax_c.plot(m_cap, -cap_change[:, 4], "b", label="discharging")
        ax_e.plot(m_cap, cap_change[:, 2], "k", label="number chg")
        ax_e.plot(m_cap, cap_change[:, 5], "b:", label="number disc")
        ax_c.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_e.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_c.set_xlabel("storage capacity")
        ax_c.set_ylabel("integral charging")
        # Raum für die Legende schaffen
        fi_n2.tight_layout(rect=[0, 0, 0.85, 1])
        fi_n2.savefig(f_name_base+"-integral-charge_cap-plot.png")
    
        # Number of full charging equivalents
        fi_n3, (ax_g, ax_h)= plt.subplots(2)
        equiv_no_charging = cap_change[:, 1] / m_cap
        equiv_no_discharging = cap_change[:, 4] / m_cap
        ax_g.plot(m_cap, equiv_no_charging, "k", label="charging")
        ax_g.plot(m_cap, equiv_no_discharging, "b:", label="discharging")
        ax_g.set_xlabel("storage capacity")
        ax_g.set_ylabel("no eq. full cyc. / y")
        ax_g.legend(loc='upper left', bbox_to_anchor=(1, 1))
    
        e2p_mean_charge = m_cap/cap_change[:, 0]
        e2p_mean_discharge = m_cap/cap_change[:, 3]
        ax_h.plot(m_cap, e2p_mean_charge, "k", label="charging")
        ax_h.plot(m_cap, e2p_mean_discharge, "b:", label="discharging")
        ax_h.set_xlabel("storage capacity")
        ax_h.set_ylabel("E2P / h")
        ax_h.legend(loc='upper left', bbox_to_anchor=(1, 1))
    
        # Raum für die Legende schaffen
        fi_n3.tight_layout(rect=[0, 0, 0.85, 1])
        fi_n3.savefig(f_name_base+"-charge_discharge-plot.png")
    
        fig_e, ax_p =plt.subplots(1)
    
        ax_p.plot(m_cap, la[:,0], "k", label="LA_pos")
        ax_p.plot(m_cap, la[:,1], "b", label="LA_neg")
        ax_p.plot(m_cap, la[:,2], "r:", label="ratio_eq")
        ax_p.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_p.set_title("Level of Autononomy (LA)")
        ax_p.set_xlabel("storage capacity")
        ax_p.set_ylabel("LA")
    
        # Raum für die Legende schaffen
        fig_e.tight_layout(rect=[0, 0, 0.85, 1])
        fig_e.savefig(f_name_base+"-la-plot.png")

    
    #-----------------------
    #-----------------------
    aa =storage(-1, 2 , 5, .5,.1666,)
    print (aa)
    pd.set_option('display.max_columns', None)
    # a, b, c = analyze_residual(data_0, 'Residuallast')
    df = pd.read_excel(DATA_DIR/'2024_Bertsch_residual_Load.xlsx')

    # Lade die Excel-Datei mit pandas

    # output_folder = create_output_folder()
    # Transponiere die Tabelle
    df_transponiert = df.set_index('node').T
    df_transponiert['time_diff_seconds'] = np.zeros(
        len(df_transponiert)) + 3600

    df_transponiert['time']= np.arange(0, len(df_transponiert) * 3600., 3600)
    countries= ["DE0 Tot"]# , "IT0 0", "GB5 0", "ES0 0", "IE4 0"]
    sum_units= 1/1000
    EFF_STOR = .75
    m_cap = np.linspace(1e6, 20e6, 15)
    for what in countries:
        fnam =f_name_base+what[:3]
        res_max = df_transponiert[what].max()
        m_cap = np.linspace(res_max/20, res_max*20, 7)

        results_df = calc_capacity_dependence(df_transponiert[what],
                                              m_cap,
                                              EFF_STOR,
                                              make_df=True,
                                              file_name=f_name_base+"-"+what[:3]+"-",
                                              title = what[:3])


        fig_e, ax_p = plt.subplots()
        line_labels = ['LA+', 'LA-', 'eq_ratio']
        colors = ['black', 'blue', 'green']  # Beispiel Farben können angepasst werden

        for label, color in zip(line_labels, colors):
            sbn.lineplot(data=results_df, x='storage capacity', y=label, ax=ax_p, label=label, color=color)

        # Anpassen der Legende
        ax_p.legend(loc='upper left', bbox_to_anchor=(1, 1))
        ax_p.set_title("Level of Autonomy (LA), "+what[:3])
        ax_p.set_xlabel("Storage Capacity")
        ax_p.set_ylabel("LA")

        # Raum für die Legende schaffen
        fig_e.tight_layout(rect=[0, 0, 0.85, 1])

        # Speichern des Plots
        fig_e.savefig("la-plot.png")

        plt.show()

        all_periods, pos_neg, neg_pos = analyze_residual(df_transponiert, what)
        neg_pos[['sum_pos', 'sum_neg']] = neg_pos[[
            'sum_pos', 'sum_neg']] *  sum_units
        sbn.scatterplot(data=neg_pos, x='count_neg', y='count_pos',
                        size = 'sum_pos', hue = 'sum_neg')  # , style='mean_pos')
        plt.legend(loc='upper left', bbox_to_anchor=(1, 1))

                            # Optional: Ändere die Größe des Plots, um Platz für die Legende zu schaffen
        plt.tight_layout()
        plt.title(what)
        plt.savefig(fnam+"sum_count.jpg")

        plt.figure()
        sbn.scatterplot(data=neg_pos, x='sum_neg', y='sum_pos',
                        size = 'count_pos', hue = 'count_neg')  # , style='mean_pos')
        plt.legend(loc='upper left', bbox_to_anchor=(1, 1))

        # Optional: Ändere die Größe des Plots, um Platz für die Legende zu schaffen
        plt.tight_layout()
        plt.title(what)
        plt.savefig(fnam+"_count_sum.jpg")
        fi, ax= plt.subplots(2)
        print(
            f"country:{what}, No.-periods: {pos_neg['count_pos'].count()}")
        act_corr= signal.correlate(
        df_transponiert[what], df_transponiert[what])
        # f, Pxx_den = signal.periodogram(data[dat_act[1][2]], fs =1/600)
        # ax4.semilogy(f, Pxx_den)
        ax[0].plot(act_corr)
        # positive_values = df_transponiert[what][df_transponiert[what] > 0]

        # Falls die Indizes wichtig sind, erstelle einen Array nur mit den positiven Werten

        # Setze negative Werte auf 0
        df_transponiert[what+"mod"] = np.where(
        df_transponiert[what] < 0, 0, df_transponiert[what])

        # Erstelle ein NumPy-Array aus der Spalte
        modified_array = df_transponiert[what+"mod"].to_numpy()

        # Berechne die Kreuzkorrelation der modifizierten Werte
        # , mode='full')
        corr = signal.correlate(modified_array, modified_array)

        # Optional: Wenn du auch die Lag-Werte berechnen möchtest
        lags = signal.correlation_lags(
        len(modified_array), len(modified_array), mode ='full')

        # Ausgabe der Korrelation und der Lags

        lc = len(corr)
        corr_n = corr[int((lc-1)/2):]
        ax[1].plot(corr_n[:200]/np.abs(corr_n).max())
        ax[0].set_title("Autocorrelation, "+what[:3])
        fi.savefig(fnam+"_correl.jpg")

        neg_pos.to_csv(fnam+"neg_pos.csv",sep=';')
        pos_neg.to_csv(fnam+"pos_neg.csv",sep=';')
        all_periods.to_csv(fnam+"all_periods.csv",sep=';')
        neg_sum = np.where(
        df_transponiert[what] > 0, 0, df_transponiert[what]).sum()
        pos_sum = np.where(
        df_transponiert[what] < 0, 0, df_transponiert[what]).sum()
        print(f"{what}, pos/neg residuals: {-pos_sum/neg_sum:.2f}")



# -*- coding: utf-8 -*-
"""
test wind data load and conversion to power

Created on Mon Jan 27 11:01:47 2025

@author: atakan
"""

import pandas as pd
import matplotlib.pyplot as plt
import zipfile
from rws_py.dwd_analyze_store_second_threshold import load_data


# Funktion zur Erzeugung synthetischer Testdaten
def create_test_data(num_hours=30):
    time_index = pd.date_range(start='2023-01-01', periods=num_hours, freq='h')
    wind_speeds = pd.Series(range(num_hours))  # Geschwindigkeit von 0 bis num_hours m/s
    data = pd.DataFrame({"MESS_DATUM": time_index, "WIND_SPEED": wind_speeds})
    return data

# Funktion zur Erstellung eines ZIP-Archivs mit der Test-CSV-Datei
def create_test_zip(data, zip_filename="testdata.zip", csv_filename="testdata.csv"):
    csv_data = data.to_csv(index=False)
    with zipfile.ZipFile(zip_filename, 'w') as zf:
        zf.writestr(csv_filename, csv_data)

# Haupttestfunktion zum Laden und Plotten der Daten
def test_load_and_plot():
    # Erstellen der Testdaten und des ZIP-Archivs
    test_data = create_test_data()
    zip_filename = "test_wind_data.zip"
    csv_filename = "test_wind_data.csv"
    create_test_zip(test_data, zip_filename, csv_filename)

    # Definition der erwarteten Spalten zum Testen
    expected_columns = ["MESS_DATUM", "WIND_SPEED"]
    directory = "."  # Verzeichnis, in dem die Test-Datei gespeichert wird

    # Laden der Daten mit der load_data-Funktion
    df = load_data(bse_url_in=None, filename_in=zip_filename,
                   exp_col=expected_columns, directory=directory,
                   energy_column=1, wind_bounds=(3.3, 11.0, 25))

   
    # Plotting
    fig, ax1 = plt.subplots(figsize=(10, 6))
    
    # Erste y-Achse für Power
    line1, = ax1.plot(df.index, df["power"], label="Power", color='b')
    ax1.set_xlabel('time')
    ax1.set_ylabel('Power (a.u.)', color='b')
    ax1.tick_params(axis='y', labelcolor='b')
    
    # Zweite y-Achse für Windgeschwindigkeit
    ax2 = ax1.twinx()
    line2, = ax2.plot(df.index, df["velocity / (m/s)"], label="Wind velocity", color='r')
    ax2.set_ylabel('Wind velocity (m/s)', color='r')
    ax2.tick_params(axis='y', labelcolor='r')
    
    # Titel hinzufügen
    plt.title('Power and Wind Velocity vs. Time')
    
    # Sammeln der Handles und Labels beider Achsen
    lines = [line1, line2]
    labels = [line.get_label() for line in lines]
    
    # Legende nach außen legen
    fig.legend(lines, labels, loc='center left', bbox_to_anchor=(1.1, 0.75))
    # Alternativ: ax1.legend(lines, labels, loc='center left', bbox_to_anchor=(1.1, 0.5))
    
    fig.tight_layout()  # um Platz für beide y-Achsen zu garantieren
    plt.show()
    
if __name__ == "__main__":
        test_load_and_plot()



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  3 15:45:31 2025

@author: atakan
"""
import os
from pathlib import Path

def append_suffix_to_path(base_path: Path, suffix: str, new_extension: str = None) -> Path:
    """
    Fügt einen Suffix zu einem Basispfad hinzu und optional eine neue Erweiterung.

    Args:
        base_path (Path): Der ursprüngliche Dateipfad.
        suffix (str): Der Suffix, der dem Dateinamen hinzugefügt werden soll (z.B. "-akt").
        new_extension (str, optional): Die neue Dateierweiterung inklusive Punkt (z.B. ".csv").
                                       Wenn None, bleibt die ursprüngliche Erweiterung erhalten.

    Returns:
        Path: Der neue Dateipfad mit hinzugefügtem Suffix und optionaler neuer Erweiterung.
    """
    # Basisname ohne Erweiterung
    base_stem = base_path.stem

    # Neuer Basisname mit Suffix
    new_stem = f"{base_stem}{suffix}"

    # Bestimmen der neuen Erweiterung
    if new_extension is not None:
        if not new_extension.startswith('.'):
            new_extension = f".{new_extension}"
        new_suffix = new_extension
    else:
        new_suffix = base_path.suffix

    # Neuer Dateiname
    new_name = f"{new_stem}{new_suffix}"

    # Neuer Pfad
    return base_path.with_name(new_name)

import pandas as pd
import rws_py

def get_stations():
    """
    Read all dwd-station numers from file 'stations_data.csv' in data folder

    Parameters
    ----------
    station_name : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    # Bestimme den Pfad zur aktuellen Datei (rws_py/__init__.py)
    current_dir = os.path.dirname(os.path.abspath(__file__))

    # Bestimme den Pfad zur Datei stations_data.csv im Ordner data
    stations_data_path = os.path.join(
        current_dir, '..', 'data', 'stations_data.csv')

    # Überprüfe, ob die Datei existiert
    if os.path.exists(stations_data_path):
        pass
        # print(f"Datei gefunden: {stations_data_path}")
    else:
        print(f"Stations file NOT found: {stations_data_path}")

    df = pd.read_csv(stations_data_path,
                     dtype=str)
    return df.set_index("Stationsname")["Stations_id"].to_dict()

if __name__ == "__main__":
    #print( get_stations())
    pass



# Beispielhafte Verwendung
if __name__ == "__main__":
    # Ursprünglicher Pfad ohne Erweiterung
    file_n = Path("C:/results/heute")
    file_aktuell1 = append_suffix_to_path(file_n, "-akt", ".csv")
    print(file_aktuell1)  # Ausgabe: C:/results/heute-akt.csv

    # Ursprünglicher Pfad mit Erweiterung
    file_mit_ext = Path("C:/results/heute.txt")
    file_aktuell2 = append_suffix_to_path(file_mit_ext, "-akt", ".csv")
    print(file_aktuell2)  # Ausgabe: C:/results/heute-akt.csv

    # Ohne Änderunge der Erweiterung
    file_aktuell3 = append_suffix_to_path(file_mit_ext, "-akt")
    print(file_aktuell3)  # Ausgabe: C:/results/heute-akt.txt



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 20 13:24:40 2025

@author: atakan
"""


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from rws_py.residual_analyse import storage


def calc_residuals(power_df,
                   load_df,
                   gamma=1.0,
                   columns={"in-power": 'power',
                            "in-load": 'load',
                            "out-residual": 'residual-no-storage',
                            "out-unchanged": 'residual-not-normalized'},
                   normalize=True):
    """
    Calculate residual load from load and power dataframes

    Parameters
    ----------
    power_df : pandas.DataFrame
        datframe with the power column.
    load_df : pandas.DataFrame
        dataframe with the load column.
    gamma : Float, optional
        the load is multiplied by this factor, before the residual is
        calculated. A measure of how much residuals are available. The default
        is 1.0.
    columns : Dictionary, optional
        where to find the input columns and in which column to write the output.
        The default is {"in-power": 'power', "in_load": 'load',
                        "out-residual": 'residual-no-storage',
                        "out-unchanged" : 'residual-not-normalized'}.
    normalize : Boolean, optional
        the residuals will be divided by the absolute mean of the negative
        residuals. The default is True.

    Returns
    -------
    power_df : pandas.DataFrame
        dataframe with the load and the residual included.

    """
    # Explizite Kopie erstellen
    result_df = power_df.copy()

    # .loc verwenden für das Zuweisen von Werten
    result_df.loc[:, columns["in-load"]] = load_df[columns['in-load']]
    result_df.loc[:, columns["out-unchanged"]] = load_df[columns['in-load']] \
        * gamma - result_df[columns['in-power']]
    result_df['$\gamma$'] = gamma

    if normalize:
        result_df.loc[:, columns["out-residual"]
                      ] = result_df[columns["out-unchanged"]]
        negative_values = result_df[result_df[columns["out-residual"]] < 0]

        mean_negative_values = abs(
            negative_values[columns["out-residual"]].sum()) \
            / result_df[columns["out-residual"]].count()
        # this is the mean per year, NOT the mean of only the times with
        # negative values!
        result_df.loc[:, columns["out-residual"]] /= mean_negative_values

    return result_df


def weather_load_storage(energy_in_df, load_df=None, storage_size=None,
                         pow_index="power",
                         eff=.75,
                         residual=False):
    """
    Calculate the remaining load and storage charging as a function of energy.

    Parameters
    ----------
    energy_in_df : pandas DataFrame
        The given index values are analyzed as time and energy.
    load_df : pandas DataFrame, optional
        If not None, a DataFrame with the date as index and load values in 
        column 'load'. The default is None.
    storage_size : float, optional
        Size in the units of "power" times hours. If None, the maximum value 
        of "power" will be multiplied by 2. The default is None.
    pow_index : string, optional
        Index of energy_df with 'power' values. The default is "power".
    eff : float, optional
        Storage efficiency (energy out / energy in). The default is 0.75.
    residual : Boolean, optional
        True, when residuals are already included in energy_df. Default: False.

    Returns
    -------
    energy_df : pandas DataFrame
        All input data frames, residuals, with and without storage,
        SOC (state of charge) of the storage.
    """
    energy_df = energy_in_df.copy()
    time_resolution_hours = energy_df.index.to_series(
    ).diff().median().total_seconds() / 3600

    if storage_size is None:
        storage_size = energy_df[pow_index].max() * 2
        print(f"St-size in: {storage_size}")

    soc_act = 0
    act_capacity = soc_act * storage_size

    col_n = ["load", "residual-no-storage", "residual-w-storage"]

    for ind in energy_df.index:
        if residual:
            res_act = energy_df.loc[ind, "residual-no-storage"]
        else:
            res_act = load_df.loc[ind, "load"] - energy_df.loc[ind, pow_index]
            energy_df.loc[ind, "load"] = load_df.loc[ind, "load"]

        st_res = storage(-res_act, act_capacity, storage_size,
                         efficiency=eff,
                         time_step=time_resolution_hours)

        energy_df.loc[ind, st_res.keys()] = st_res.values()
        energy_df.loc[ind, "residual-no-storage"] = res_act
        energy_df.loc[ind, "residual-w-storage"] = res_act + \
            st_res["power_st_in"]

        act_capacity = st_res["act_st_cap"]

    return energy_df


def test_weather_load_storage(verbose=False, hours=48):
    # np.random.seed(42)
    dates = pd.date_range("2023-01-01", periods=hours, freq="h")

    # Create situations with different load and energy patterns
    energy_values = np.random.choice(
        [0] + list(np.arange(00, 1001.)), size=len(dates))
    load_values_high = np.linspace(
        900, -150, num=len(dates))  # Load greater than energy
    # Load less than energy
    load_values_low = np.linspace(-300, 950, num=len(dates))
    load_values_equal = energy_values                          # Load equal to energy
    load_values_mean_equal = np.random.choice(
        [0] + list(np.arange(00, 1001.)), size=len(dates))

    energy_df = pd.DataFrame(
        {"MESS_DATUM": dates, "power": energy_values}, index=dates)
    load_df_high = pd.DataFrame({"load": load_values_high}, index=dates)
    load_df_low = pd.DataFrame({"load": load_values_low}, index=dates)
    load_df_equal = pd.DataFrame({"load": load_values_equal}, index=dates)
    load_mean_equal = pd.DataFrame(
        {"load": load_values_mean_equal}, index=dates)

    # Call the function for each scenario
    scenarios = {"High to low Load": load_df_high,
                 "Low Load": load_df_low,
                 "Equal Load": load_df_equal,
                 "Mean Load equal": load_mean_equal}

    for scenario, load_df in scenarios.items():
        test_df = calc_residuals(energy_df, load_df)
        print('\nFrom calc_residuals', test_df.describe())
        result_df = weather_load_storage(energy_df.copy(), load_df)

        if verbose:
            plt.figure(figsize=(10, 4))
            plt.plot(
                result_df.index, result_df['residual-w-storage'], "<g", label='Residual with Storage')
            plt.plot(result_df.index, result_df['residual-no-storage'],
                     "vb", label='Residual without Storage', linestyle='--')
            plt.plot(result_df.index,
                     result_df['load'], "ok", label='Load', linestyle='-.')
            plt.plot(
                result_df.index, result_df['act_st_cap'], "k", label='act_st_cap', linestyle='-.')
            plt.plot(result_df.index,
                     energy_df['power'], "k", label='power', linestyle='-')
            plt.xlabel('Time')
            plt.ylabel('power / [a.u.]')
            plt.title(f'Scenario: {scenario}')
            plt.legend(loc='center left', bbox_to_anchor=(1.1, 0.75))
            # Alternativ: ax1.legend(lines, labels, loc='center left', bbox_to_anchor=(1.1, 0.5))

            plt.tight_layout()  # um Platz für beide y-Achsen zu garantieren
            plt.grid(True)
            plt.show()
            print(f"\n{scenario}")
            print(result_df.describe())
            print_summary_stats(result_df, energy_df)


def print_summary_stats(result_df, energy_df):
    metrics = {
        'Residual with Storage': result_df['residual-w-storage'],
        'Residual without Storage': result_df['residual-no-storage'],
        'Load': result_df['load'],
        'Energy': energy_df['power']
    }

    print("{:<25} {:>10} {:>10} {:>10}".format(
        "Kategorie", "Mittelwert", "Minimum", "Maximum"))
    print("-" * 57)

    for label, data in metrics.items():
        mean_val = data.mean()
        min_val = data.min()
        max_val = data.max()
        print(f"{label:<25} {mean_val:>10.2f} {min_val:>10.2f} {max_val:>10.2f}")

# Beispiel, um die Funktion zu verwenden:
# print_summary_stats(result_df, energy_df)


if __name__ == "__main__":
    test_weather_load_storage(verbose=True)
